/*
 * arch/ppc/platforms/rsdproto_time.c
 * 
 * (C) Copyright 2000 Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 */

#include <linux/types.h>
#include <asm/delay.h>
#include <asm/immap_8260.h>
#include <asm/pgtable.h>
#include <asm/mpc8260.h>
#include <asm/cpm_8260.h>
#include <asm/time.h>

/* define to enable debug messages */
#undef DEBUG_STEP

/* us to wait before checking the i2c */
#define DELAY_US    100000  

#define I2C_RXTX_LEN 128 /* maximum tx/rx buffer length */
#define NUM_RX_BDS 4
#define NUM_TX_BDS 4
#define MAX_TX_SPACE 256
#define MAX_RX_SPACE 256

typedef struct I2C_BD
{
  unsigned short status;
  unsigned short length;
  unsigned char *addr;
} I2C_BD;
#define BD_I2C_TX_START 0x0400	/* special status for i2c: Start condition */

#ifdef DEBUG_STEP 
#define PRINTD(x) printk x
#else
#define PRINTD(x)
#endif

extern unsigned char __res[];

/* This structure keeps track of the bd and buffer space usage. */
typedef struct i2c_state {
	int rx_idx, tx_idx;		/* index to next free rx/tx bd */
	void *rxbd, *txbd;		/* pointer to next free rx/tx bd */
	int tx_space;			/* number of tx bytes left */
	unsigned char *tx_buf;	/* pointer to free tx area */
	int rx_space;			/* number of rx bytes left */
	unsigned char *rx_buf;	/* pointer to free rx area */
} i2c_state_t;

/* flags for i2c_send() and i2c_receive() */
#define I2CF_ENABLE_SECONDARY	0x01	/* secondary_address is valid */
#define I2CF_START_COND			0x02	/* tx: generate start condition */
#define I2CF_STOP_COND			0x04	/* tx: generate stop condition */

/* return codes */
#define I2CERR_NO_BUFFERS		0x01	/* no more bds or buffer space */
#define I2CERR_MSG_TOO_LONG		0x02	/* tried to send/receive to much data */

/*
 * Returns the best value of I2BRG to meet desired clock speed of I2C with
 * input parameters (clock speed, filter, and predivider value).
 * It returns computer speed value and the difference between it and desired
 * speed.
 */
static inline int i2c_roundrate(int hz, int speed, int filter, int modval,
				int *brgval, int *totspeed)
{
    int moddiv = 1 << (5-(modval & 3)), brgdiv, div;
    
    brgdiv = hz / (moddiv * speed);

    *brgval = brgdiv / 2 - 3 - 2*filter;

    if ((*brgval < 0) || (*brgval > 255))
      return -1;

    brgdiv = 2 * (*brgval + 3 + 2 * filter);
    div = moddiv * brgdiv ;
    *totspeed = hz / div;

    return  0;
}

/* 
 * Sets the I2C clock predivider and divider to meet required clock speed.
 */
static int i2c_setrate(int hz, int speed)
{
    immap_t *immap = (immap_t *)IMAP_ADDR;
    volatile i2c8260_t *i2c = (i2c8260_t *)&immap->im_i2c;
    int brgval,
      modval,	/* 0-3 */
      bestspeed_diff = speed,
      bestspeed_brgval=0,
      bestspeed_modval=0,
      bestspeed_filter=0,
      totspeed,
      filter = 0; /* Use this fixed value */
    
    for (modval = 0; modval < 4; modval++)
    {
	if (i2c_roundrate (hz, speed, filter, modval, &brgval, &totspeed) == 0)
	{
	    int diff = speed - totspeed ;

	    if ((diff >= 0) && (diff < bestspeed_diff))
	    {
		bestspeed_diff	= diff ;
		bestspeed_modval    = modval;
		bestspeed_brgval    = brgval;
		bestspeed_filter    = filter;
	    }
	}
    }
    
    PRINTD(("[I2C] Best is:\n"));
    PRINTD(("[I2C] CPU=%dhz RATE=%d F=%d I2MOD=%08x I2BRG=%08x DIFF=%dhz\n",
	   hz, speed,
	   bestspeed_filter, bestspeed_modval, bestspeed_brgval,
	   bestspeed_diff));

    i2c->i2c_i2mod |= ((bestspeed_modval & 3) << 1) | (bestspeed_filter << 3);
    i2c->i2c_i2brg = bestspeed_brgval & 0xff;
    
    PRINTD(("[I2C] i2mod=%08x i2brg=%08x\n", i2c->i2c_i2mod, i2c->i2c_i2brg));
    
    return 1 ;
}

static void i2c_init(int speed, int slaveaddr)
{
    bd_t *binfo = (bd_t *)__res;
    volatile immap_t *immap = (immap_t *)IMAP_ADDR;
    volatile cpm8260_t *cp = (cpm8260_t *)&immap->im_cpm;
    volatile i2c8260_t *i2c = (i2c8260_t *)&immap->im_i2c;
    volatile iic_t *iip;
    ulong rbase, tbase;
    volatile I2C_BD *rxbd, *txbd;

    /*
     * initialise data in dual port ram:
     * 
     * BD_IIC_START  -> parameter ram (64 bytes)
     *	       rbase -> rx BD	      (NUM_RX_BDS * sizeof(I2C_BD) bytes)
     *	       tbase -> tx BD	      (NUM_TX_BDS * sizeof(I2C_BD) bytes)
     *			tx buffer     (MAX_TX_SPACE bytes)
     *			rx buffer     (MAX_RX_SPACE bytes)
     */

    /* locate i2c parameter ram at BD_IIC_START */
    iip = (iic_t *)&immap->im_dprambase[BD_IIC_START];
    *((unsigned short*)(&immap->im_dprambase[PROFF_I2C_BASE])) = BD_IIC_START;

    memset((void*)iip, 0, sizeof(iic_t));

    rbase = BD_IIC_START+64;
    tbase = rbase + NUM_RX_BDS * sizeof(I2C_BD);

    /* Disable interrupts */
    i2c->i2c_i2mod = 0x00;
    i2c->i2c_i2cmr = 0x00;
    i2c->i2c_i2cer = 0xff;
    i2c->i2c_i2add = slaveaddr;
    
    /*
     * Set the I2C BRG Clock division factor from desired i2c rate
     * and current CPU rate (we assume sccr dfbgr field is 0;
     * divide BRGCLK by 1)
     */ 
    PRINTD(("[I2C] Setting rate...\n"));
    i2c_setrate (binfo->bi_brgfreq, speed) ;
    
    /* Set I2C controller in master mode */
    i2c->i2c_i2com = 0x01;

    /* Initialize Tx/Rx parameters */
    iip->iic_rbase = rbase;
    iip->iic_tbase = tbase;
    rxbd = (I2C_BD *)((unsigned char *)&immap->im_dprambase[iip->iic_rbase]);
    txbd = (I2C_BD *)((unsigned char *)&immap->im_dprambase[iip->iic_tbase]);
    
    PRINTD(("[I2C] rbase = %04x\n", iip->iic_rbase));
    PRINTD(("[I2C] tbase = %04x\n", iip->iic_tbase));
    PRINTD(("[I2C] rxbd = %08x\n", (int)rxbd));
    PRINTD(("[I2C] txbd = %08x\n", (int)txbd));

    /* Set big endian byte order */
    iip->iic_tfcr = 0x10;
    iip->iic_rfcr = 0x10;

    /* Set maximum receive size. */
    iip->iic_mrblr = I2C_RXTX_LEN;
    
    cp->cp_cpcr = mk_cr_cmd(CPM_CR_I2C_PAGE,
			    CPM_CR_I2C_SBLOCK, 
			    0x00,
			    CPM_CR_INIT_TRX) | CPM_CR_FLG;
    do {
	__asm__ __volatile__ ("eieio");
    } while (cp->cp_cpcr & CPM_CR_FLG);

    /* Clear events and interrupts */
    i2c->i2c_i2cer = 0xff;
    i2c->i2c_i2cmr = 0x00;
}

void i2c_newio(i2c_state_t *state)
{
    volatile immap_t *immap = (immap_t *)IMAP_ADDR;
    volatile iic_t *iip;

    PRINTD(("[I2C] i2c_newio\n"));
    
    iip = (iic_t *)&immap->im_dprambase[BD_IIC_START];
    state->rx_idx = 0;
    state->tx_idx = 0;
    state->rxbd = (void*)&immap->im_dprambase[iip->iic_rbase];
    state->txbd = (void*)&immap->im_dprambase[iip->iic_tbase];
    state->tx_space = MAX_TX_SPACE;
    state->tx_buf = (unsigned char*)state->txbd + NUM_TX_BDS * sizeof(I2C_BD);
    state->rx_space = MAX_RX_SPACE;
    state->rx_buf = state->tx_buf + MAX_TX_SPACE;

    PRINTD(("[I2C] rxbd = %08x\n", (int)state->rxbd));
    PRINTD(("[I2C] txbd = %08x\n", (int)state->txbd));
    PRINTD(("[I2C] tx_buf = %08x\n", (int)state->tx_buf));

    /* clear the buffer memory */
    memset((char *)state->tx_buf, 0, MAX_TX_SPACE);
}

int i2c_send(i2c_state_t *state,
	     unsigned char address,
	     unsigned char secondary_address,
	     unsigned int flags,
	     unsigned short size, 
	     unsigned char *dataout)
{
    volatile I2C_BD *txbd;
    int i,j;

    PRINTD(("[I2C] i2c_send add=%02d sec=%02d flag=%02d size=%d\n", 
	    address, secondary_address, flags, size));
    
    /* trying to send message larger than BD */
    if (size > I2C_RXTX_LEN)
      return I2CERR_MSG_TOO_LONG;

    /* no more free bds */
    if (state->tx_idx >= NUM_TX_BDS || state->tx_space < (2 + size))
      return I2CERR_NO_BUFFERS;
    
    txbd = (I2C_BD *)state->txbd;
    txbd->addr = state->tx_buf;

    PRINTD(("[I2C] txbd = %08x\n", (int)txbd));

    PRINTD(("[I2C] Formatting addresses...\n"));
    if (flags & I2CF_ENABLE_SECONDARY)
    {
	txbd->length = size + 2;  /* Length of message plus dest addresses */
	txbd->addr[0] = address << 1;
	txbd->addr[0] &= ~(0x01);
	txbd->addr[1] = secondary_address;
	i = 2;
    }
    else
    {
	txbd->length = size + 1;  /* Length of message plus dest address */
	txbd->addr[0] = address << 1;  /* Write destination address to BD */
	txbd->addr[0] &= ~(0x01);  /* Set address to write */
	i = 1;
    }

    /* set up txbd */
    txbd->status = BD_SC_READY;
    if (flags & I2CF_START_COND)
      txbd->status |= BD_I2C_TX_START;
    if (flags & I2CF_STOP_COND)
      txbd->status |= BD_SC_LAST | BD_SC_WRAP;
    
    /* Copy data to send into buffer */
    PRINTD(("[I2C] copy data...\n"));
    for(j = 0; j < size; i++, j++)
      txbd->addr[i] = dataout[j];

    PRINTD(("[I2C] txbd: length=0x%04x status=0x%04x addr[0]=0x%02x addr[1]=0x%02x\n",
	   txbd->length,
	   txbd->status,
	   txbd->addr[0],
	   txbd->addr[1]));

    /* advance state */
    state->tx_buf += txbd->length;
    state->tx_space -= txbd->length;
    state->tx_idx++;
    state->txbd = (void*)(txbd + 1);

    return 0;
}

int i2c_receive(i2c_state_t *state,
		unsigned char address,
		unsigned char secondary_address,
		unsigned int flags,
		unsigned short size_to_expect, 
		unsigned char **datain)
{
    volatile I2C_BD *rxbd, *txbd;

    PRINTD(("[I2C] i2c_receive %02d %02d %02d\n", address, secondary_address, flags));

    /* Expected to receive too much */
    if (size_to_expect > I2C_RXTX_LEN)
      return I2CERR_MSG_TOO_LONG;
    
    /* no more free bds */
    if (state->tx_idx >= NUM_TX_BDS || state->rx_idx >= NUM_RX_BDS
	 || state->tx_space < 2 || size_to_expect > state->rx_space)
      return I2CERR_NO_BUFFERS;

    rxbd = (I2C_BD *)state->rxbd;
    txbd = (I2C_BD *)state->txbd;

    PRINTD(("[I2C] rxbd = %08x\n", (int)rxbd));
    PRINTD(("[I2C] txbd = %08x\n", (int)txbd));

    /* set up txbd */
    txbd->addr = state->tx_buf;
    if (flags & I2CF_ENABLE_SECONDARY)
    {
	txbd->length = 2;
	txbd->addr[0] = address << 1;	/* Write data */
	txbd->addr[1] = secondary_address;  /* Internal address */
	txbd->status = BD_SC_READY;
    }
    else
    {
	txbd->length = 1 + size_to_expect;
	txbd->addr[0] = (address << 1) | 0x01;
	txbd->status = BD_SC_READY;
	memset(&txbd->addr[1], 0, txbd->length);
    }

    /* set up rxbd for reception */
    rxbd->status = BD_SC_EMPTY;
    rxbd->length = size_to_expect;
    *datain = rxbd->addr = state->rx_buf;

    if (flags & I2CF_START_COND)
      txbd->status |= BD_I2C_TX_START;
    if (flags & I2CF_STOP_COND)
    {
	txbd->status |= BD_SC_LAST | BD_SC_WRAP;
	rxbd->status |= BD_SC_WRAP;
    }

    PRINTD(("[I2C] txbd: length=0x%04x status=0x%04x addr(phys)=0x%08lx\n",
	   txbd->length,
	   txbd->status,
	   txbd->addr));
    PRINTD(("[I2C] rxbd: length=0x%04x status=0x%04x addr(phys)=0x%08lx\n",
	   rxbd->length,
	   rxbd->status,
	   rxbd->addr));

    /* advance state */
    state->tx_buf += txbd->length;
    state->tx_space -= txbd->length;
    state->rx_buf += rxbd->length;
    state->rx_space -= rxbd->length;
    state->tx_idx++;
    state->txbd = (void*)(txbd + 1);
    state->rx_idx++;
    state->rxbd = (void*)(rxbd + 1);

    return 0;
}


int i2c_doio(i2c_state_t *state)
{
    volatile immap_t *immap = (immap_t *)IMAP_ADDR;
    volatile iic_t *iip;
    volatile i2c8260_t *i2c = (i2c8260_t *)&immap->im_i2c;
    volatile I2C_BD *txbd, *rxbd;

    PRINTD(("[I2C] i2c_doio\n"));

    txbd = ((I2C_BD*)state->txbd) - 1;
    rxbd = ((I2C_BD*)state->rxbd) - 1;

    /* make sure txbd/rxbd terminate the chain */
    txbd->status |= BD_SC_WRAP;
    rxbd->status |= BD_SC_WRAP;

    iip = (iic_t *)&immap->im_dprambase[BD_IIC_START];
    iip->iic_rbptr = iip->iic_rbase;
    iip->iic_tbptr = iip->iic_tbase;

    /* Enable I2C */
    PRINTD(("[I2C] Enabling I2C...\n"));
    i2c->i2c_i2mod |= 0x01;

    /* Begin transmission */
    i2c->i2c_i2com |= 0x80;

    /* This is a patch! */
    udelay (DELAY_US) 
      ; 

    /* Loop until transmit & receive completed */
    if (state->tx_idx > 0) 
    {
	PRINTD(("[I2C] Transmitting...\n"));
	while(txbd->status & BD_SC_READY)   
	  __asm__ __volatile__ ("eieio");
    }
    if (state->rx_idx > 0) 
    {
	PRINTD(("[I2C] Receiving...\n"));
	while(rxbd->status & BD_SC_EMPTY)
	  __asm__ __volatile__ ("eieio");
    }
    
    /* Turn off I2C */
    i2c->i2c_i2mod &= ~0x01;
    
    return 0;
}

/* I2C Bus adresses for PPC & Protocol board */
#define PPC8260_I2C_ADR	    0x30
#define RS5C372_PPC_I2C_ADR 0x32

#ifndef BCD_TO_BIN
#define BCD_TO_BIN(val) ((val)=((val)&15) + ((val)>>4)*10)
#endif

#ifndef BIN_TO_BCD
#define BIN_TO_BCD(val) ((val)=(((val)/10)<<4) + (val)%10)
#endif

static int rtc_displays_12 = 0; /* rtc uses 12hours (AM/PM) display format */

/* interface to linux */
int rsdproto_set_rtc_time(unsigned long nowtime)
{
    struct rtc_time tm;
    i2c_state_t i2c_state;
    unsigned char buffer[8];
    int rc;

    /* initialise i2c */
    i2c_init(100000 >> 1, PPC8260_I2C_ADR);
      
    to_tm(nowtime, &tm);

    /* handle year 2000 problem */
    tm.tm_year = (tm.tm_year - 1900) % 100;

    /* fix 12/24 display format */
	if (rtc_displays_12 && tm.tm_hour > 12)
      tm.tm_hour = tm.tm_hour + 20 - 12;

    buffer[0] = BIN_TO_BCD(tm.tm_sec);
    buffer[1] = BIN_TO_BCD(tm.tm_min);
    buffer[2] = BIN_TO_BCD(tm.tm_hour);
    buffer[3] = BIN_TO_BCD(tm.tm_wday);
    buffer[4] = BIN_TO_BCD(tm.tm_mday);
    buffer[5] = BIN_TO_BCD(tm.tm_mon);
    buffer[6] = BIN_TO_BCD(tm.tm_year);

    i2c_newio(&i2c_state);
    
    /* schedule send command with start condition */
    rc = i2c_send(&i2c_state,
		 RS5C372_PPC_I2C_ADR,	    /* address */
		 0x00,			    /* secondary address */
		 I2CF_ENABLE_SECONDARY | I2CF_START_COND | I2CF_STOP_COND,
		 7,			    /* size */
		 buffer);		    /* data */
    if (rc) 
      panic("i2c error %02x\n", rc);

    /* perform io operations */
    rc = i2c_doio(&i2c_state);
    if (rc) 
      panic("i2c error %02x\n", rc);

    return 0;
}

/* interface to linux */
unsigned long rsdproto_get_rtc_time(void)
{
    unsigned int year, mon, day, hour, min, sec;
    i2c_state_t i2c_state;
    unsigned char *buffer = 0;
    int rc;

    /* initialise i2c */
    i2c_init(100000 >> 1, PPC8260_I2C_ADR);

    i2c_newio(&i2c_state);
    
    /* schedule send command with start condition */
    rc = i2c_send(&i2c_state,
		 RS5C372_PPC_I2C_ADR,	    /* address */
		 0x00,			    /* secondary address */
		 I2CF_ENABLE_SECONDARY | I2CF_START_COND,
		 0,			    /* size */
		 NULL);			    /* data */
    if (rc) 
      panic("i2c error %02x\n", rc);

    /* schedule the read command with repeated start condition */
    rc = i2c_receive(&i2c_state,
		    RS5C372_PPC_I2C_ADR,	/* address */
		    0,				/* secondary address */
		    I2CF_START_COND | I2CF_STOP_COND,
		    16,				/* size to expect */
		    &buffer);			/* data */
    if (rc) 
      panic("i2c error %02x\n", rc);

    /* perform io operations */
    rc = i2c_doio(&i2c_state);
    if (rc) 
      panic("i2c error %02x\n", rc);

    sec	 = BCD_TO_BIN(buffer[0]);
    min	 = BCD_TO_BIN(buffer[1]);
    hour = BCD_TO_BIN(buffer[2]);
    day	 = BCD_TO_BIN(buffer[4]);
    mon	 = BCD_TO_BIN(buffer[5]);
    year = BCD_TO_BIN(buffer[6]);

    /* fix 12/24 display format */
    rtc_displays_12 = !(buffer[15] & 32);
	if (rtc_displays_12 && hour >= 30)
      hour = hour - 20 + 12;

    /* handle year 2000 problem */
    if ((year += 1900) < 1970)
	year += 100;
#if 0	
    printk("%04d-%02d-%02d %02d:%02d:%02d.\n", year, mon, day, hour, min, sec);
#endif
    return mktime(year, mon, day, hour, min, sec);
}
