/*********************  P r o g r a m  -  M o d u l e ***********************
 *  
 *         Name: vme_menpci2vme.c
 *      Project: Linux for MEN 824x ALI boards
 *
 *       Author: kp
 *        $Date: $
 *    $Revision: $
 *
 *  Description: driver for MEN PCI to VME bridge that can be found on 
 *				 MEN B11 or MEN A12
 *                      
 *                      
 *     Required: -
 *     Switches: -
 *
 *-------------------------------[ History ]---------------------------------
 *
 * $Log: $
 *---------------------------------------------------------------------------
 * (c) Copyright 2002 by MEN mikro elektronik GmbH, Nuernberg, Germany 
 ****************************************************************************/
 
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>

#include <linux/kernel.h> /* printk() */
#include <linux/slab.h> /* kmalloc() */
#include <linux/vmalloc.h> /* vmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/pci.h>
#include <linux/init.h>

#include <linux/men_vme_kernelif.h>

/*--------------------------------------+
|   DEFINES                             |
+--------------------------------------*/
MODULE_AUTHOR("Klaus Popp <klaus.popp@men.de>");
MODULE_DESCRIPTION("MEN PCI to VME bridge driver");
MODULE_LICENSE("GPL");

#define DEBUG
#ifdef DEBUG
# define DBG(x...) printk(x)
#else
# define DBG(x...)
#endif

/*--- PCI vendor/device IDs ---*/
#define MENPCI2VME_DEV_ID		0x5056	/* ASCII 'PV' */
#define MENPCI2VME_VEN_ID		0x1172  

/*--- offsets to address found in BAR0 ---*/
#define MENPCI2VME_A24_SPACE	0x00000000	/* VME A24 master access */
#define MENPCI2VME_A16_SPACE	0x01000000	/* VME A16 master access */
#define MENPCI2VME_SRAM_SPACE	0x01400000	/* VME shared RAM */
#define MENPCI2VME_REG_SPACE	0x01800000 	/* registers */
#define MENPCI2VME_IACK_SPACE	0x01c00000 	/* IACK space */

/*--- offsets to register start address (MENPCI2VME_REG_SPACE) ---*/
#define MENPCI2VME_INTR			0x00	/* interrupter control */
#define MENPCI2VME_INTID		0x04	/* interrupter vector */
#define MENPCI2VME_ISTAT		0x08	/* interrupt pending */
#define MENPCI2VME_IMASK		0x0c	/* interrupt handler mask*/
#define MENPCI2VME_MSTR			0x10	/* master control */
#define MENPCI2VME_SLV			0x14	/* slave control */
#define MENPCI2VME_SYSCTL		0x18	/* system control */

/*--- bit defs of registers above ---*/
#define MENPCI2VME_INTR_INTEN	0x8

#define MENPCI2VME_ISTAT_ACFST	0x1

#define MENPCI2VME_IMASK_ACFEN	0x1

#define MENPCI2VME_MSTR_BERR	0x4
#define MENPCI2VME_MSTR_REQ	0x2
#define MENPCI2VME_MSTR_RMW	0x1

#define MENPCI2VME_SLV_SLEN	0x10

#define MENPCI2VME_SYSCTL_ATO		0x4
#define MENPCI2VME_SYSCTL_SYSRES 	0x2
#define MENPCI2VME_SYSCTL_SYSCON 	0x1

#define ACTION_HANDLER 	0x1
#define ACTION_USERPROC 0x2

#define VME_REG_READ(reg)	(readb((char *)G_vRegs + (reg)))
#define VME_REG_WRITE(reg, val)	(writeb(val, (char *)G_vRegs + (reg)))

#define VME_CTRL_SETMASK(reg,mask)  \
 VME_REG_WRITE((reg),VME_REG_READ(reg)|(mask))
#define VME_CTRL_CLRMASK(reg,mask)	\
 VME_REG_WRITE((reg),VME_REG_READ(reg)&~(mask))

/*--------------------------------------+
|   TYPDEFS                             |
+--------------------------------------*/
/* structure to maintain installed irqs */
struct vme_irq_entry {
	struct vme_irq_entry *next;	/* ptr to next entry using same vector */

	int action;					/* ACTION_HANDLER and/or ACTION_USERPROC */

    /* for ACTION_USERPROC */
	struct task_struct *task;	/* user task structure to signal */

	/* for ACTION_HANDLER */
	const char *device;			/* kernel mode device name */
	void (*handler)(int, void *, struct pt_regs *);	/* handler */
	void *dev_id;				/* handler's private data */
};

/*--------------------------------------+
|   EXTERNALS                           |
+--------------------------------------*/

/*--------------------------------------+
|   GLOBALS                             |
+--------------------------------------*/
static int G_nBridges;			/* number of bridges found */
static u32 G_physBase;			/* phys address from BAR0 */
static void *G_vIack, *G_vRegs;	/* virtual address of registers and IACK */
static struct vme_irq_entry *G_irqTbl[256];

/*--------------------------------------+
|   PROTOTYPES                          |
+--------------------------------------*/
static void menpci2vme_irq_handler(int irq, void *dev_id, 
								   struct pt_regs *regs);

EXPORT_SYMBOL(vme_request_irq);
EXPORT_SYMBOL(vme_free_irq);
EXPORT_SYMBOL(vme_bus_to_phys);
EXPORT_SYMBOL(vme_ilevel_control);



/********************************* menpci2vme_probe **************************
 *
 *  Description: Probe/initialize on VME bridge
 *
 * gets called (during execution of pci_register_driver for already
 * existing devices or later if a new device gets inserted) for all
 * PCI devices which match the ID table and are not handled by the
 * other drivers yet. This function gets passed a pointer to the
 * pci_dev structure representing the device and also which entry in
 * the ID table did the device match. It returns zero when the driver
 * has accepted the device or an error code (negative number)
 * otherwise.
 *			   
 *---------------------------------------------------------------------------
 *  Input......: pdev		pci_dev structure representing the device
 *				 ent		entry in PCI table thta matches
 *  Output.....: returns 	zero when the driver has accepted the device or 
 *							an error code (negative number) otherwise.
 *  Globals....: -
 ****************************************************************************/
static int __devinit menpci2vme_init_one (
	struct pci_dev *pdev,
	const struct pci_device_id *ent)
{
	if( ++G_nBridges > 1 ){
		DBG("menpci2vme: cannot handle more than one bridge\n");
		return -EBUSY;			/* can handle only one bridge */
	}

	printk( "menpci2vme: found bridge (Device number %d)\n", 
			PCI_SLOT( pdev->devfn ));

	/* get base address of all the bridge space */
	G_physBase = pci_resource_start(pdev, 0);

	DBG(" physBase=0x%x\n", G_physBase );

	/* reserve registers and IACK space */
	if (!request_mem_region( G_physBase + MENPCI2VME_REG_SPACE, 0x800000, 
							 "menpci2vme")) {

		printk (KERN_ERR "menpci2vme: cannot reserve MMIO region\n");
		goto err0;
	}
	
	/* map 4k for registers and 4k for IACK */
	G_vRegs = ioremap_nocache( G_physBase + MENPCI2VME_REG_SPACE, 0x1000 );
	if( !G_vRegs ){
		printk (KERN_ERR "menpci2vme: cannot ioremap regs\n");
		goto err1;
	}
		
	G_vIack = ioremap_nocache( G_physBase + MENPCI2VME_IACK_SPACE, 0x1000 );
	if( !G_vIack ){
		printk (KERN_ERR "menpci2vme: cannot ioremap IACK space\n");
		goto err2;
	}
	
	if (pci_enable_device(pdev))
		goto err3;
	
	/* request interrupt */
	if( request_irq( pdev->irq, menpci2vme_irq_handler, SA_SHIRQ, 
					 "vme-menpci2vme", NULL )){
		printk (KERN_ERR "menpci2vme: cannot attach to irq %d\n", pdev->irq);
		goto err3;
	}
	return 0;
 err3:
	iounmap( G_vIack );
 err2:
	iounmap( G_vRegs );
 err1:
	release_mem_region( G_physBase + MENPCI2VME_REG_SPACE, 0x800000 );
 err0:	
	return -ENODEV;
}

static void __devexit menpci2vme_remove_one (struct pci_dev *pdev)
{
	free_irq( pdev->irq, NULL );
	iounmap( G_vIack );
	iounmap( G_vRegs );
	release_mem_region( G_physBase + MENPCI2VME_REG_SPACE, 0x800000 );
}	

/********************************* menpci2vme_irq_handler ********************
 *
 *  Description: VME Interrupt handler
 *			   
 *			   
 *---------------------------------------------------------------------------
 *  Input......: 
 *  Output.....: 
 *  Globals....: -
 ****************************************************************************/
static void menpci2vme_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
  int vector;
  unsigned char istat;
  int level = 0;

  /*--- get pending interrupt levels. Only unmasked levels will show up ---*/
  istat = VME_REG_READ( MENPCI2VME_ISTAT );eieio();
        
  if( istat == 0 )
	return;					/* no interrupt pending */
	
  /*--- decode level ---*/
  if( istat & 0x80 ) 	  level = 7; 
  else if( istat & 0x40 ) level = 6;
  else if( istat & 0x20 ) level = 5;
  else if( istat & 0x10 ) level = 4;
  else if( istat & 0x08 ) level = 3;
  else if( istat & 0x04 ) level = 2;
  else if( istat & 0x02 ) level = 1;
  else if( istat & 0x01 ) {
	/* ACFAIL interrupt */
	vector = 0;
	VME_CTRL_SETMASK( MENPCI2VME_ISTAT, MENPCI2VME_ISTAT_ACFST );
 	return;
  } 
  else return;    


  /*--- Get the interrupt vector from IACK space ---*/

  /* clear bus error bit to detect spurious interrupts */
  VME_CTRL_SETMASK( MENPCI2VME_MSTR, MENPCI2VME_MSTR_BERR );

  /* fetch vector */
  vector = readb( (char *)G_vIack + (level<<1) + 1 );
  
  /*printk("Vme driver irq handler, irq level %d vector %d\n",level, vector);*/

  if( VME_REG_READ( MENPCI2VME_MSTR ) & MENPCI2VME_MSTR_BERR )
	  printk( KERN_WARNING "Spurious VME Interrupt! level %d vector %d\n", 
			  level, vector);
  else {  
	  struct vme_irq_entry *irqent = G_irqTbl[vector];

	  if( !irqent ){
		  printk( KERN_WARNING "Uninitialized VME Interrupt! "
				  "level %d vector %d\n", level, vector);
	  }
		  
	  /* now dispatch to connected routines */
	  while( irqent ){
		  if((irqent->action & ACTION_HANDLER) && irqent->handler ){
			  /* call kernel mode handler */
			  irqent->handler( vector, irqent->dev_id, regs );
		  }

#if 0							/* ??? */
		  if ((irqent->action & ACTION_USERPROC) && irqent->task)
		  {
			  send_sig(SIGVME,irqent->task,1);
		  }
#endif
		  irqent = irqent->next;
	  }

	}	      
  
}

/********************************* vme_bus_to_phys ***************************
 *
 *  Description: Convert VME address to CPU physical address
 *
 *	The returned address will be ioremap'ed by the calling driver
 *---------------------------------------------------------------------------
 *  Input......: space		VME space to use (VME_XXX_SPACE) or'ed with
 *							data width (VME_XXX_ACCESS)
 *				 vmeadrs	offset in VME space
 *  Output.....: returns:	Linux error number (negative)
 *				 *physadrs_p CPU physical address
 *  Globals....:
 ****************************************************************************/
int vme_bus_to_phys( int space, u32 vmeadrs, void **physadrs_p )
{
	switch (space)
    {
	case VME_A16_SPACE|VME_D16_ACCESS:
		if( vmeadrs >= 0x10000 )
			return -EINVAL;

		*physadrs_p = (void *)(G_physBase + MENPCI2VME_A16_SPACE + vmeadrs);
		break;
		
	case VME_A24_SPACE|VME_D16_ACCESS:
		if( vmeadrs >= 0x1000000 )
			return -EINVAL;

		*physadrs_p = (void *)(G_physBase + MENPCI2VME_A24_SPACE + vmeadrs);
		break;
	default:
		return -ENODEV;
	}
	return 0;
}

/********************************* vme_request_irq ***************************
 *
 *  Description: Install handler for VME interrupt
 *			   
 *  This implementation allows multiple handlers to be installed on 
 *  the same vector
 *
 *  This call does not implicitely enable interrupts. 
 *  See vme_ilevel_control()  
 *---------------------------------------------------------------------------
 *  Input......: vme_irq		VME irq number to install
 *				 handler		handler to call 
 *				 flags			(currently not used)
 *				 device			device name registered for this irq
 *				 dev_id			ptr that is passed to handler
 *  Output.....: returns:		Linux error number (negative)
 *  Globals....:
 ****************************************************************************/
int vme_request_irq(
	unsigned int vme_irq,
	void (*handler)(int, void *, struct pt_regs *),
	unsigned long flags,
	const char *device, 
	void *dev_id)
{
	unsigned long _flags;
	struct vme_irq_entry *irqent;
	//printk("vme_request_irq: irq %d device %s\n", vme_irq, device );

	if( vme_irq >= 256 ) 
		return -EINVAL;

	irqent = kmalloc( sizeof( *irqent ), GFP_KERNEL );
	if( irqent == 0 ){
		printk(KERN_WARNING  "vme_request_irq: "
			   "Cannot allocate irqent struct!\n");
		return -ENOMEM;
	}
	memset( irqent, 0, sizeof(*irqent));

	irqent->action  = ACTION_HANDLER;
	irqent->handler = handler;
	irqent->dev_id	= dev_id;
	irqent->device	= device;

	irqent->next	= G_irqTbl[vme_irq];

	/* install it as first handler */
	__save_flags(_flags);
	cli();

	G_irqTbl[vme_irq] = irqent;

	__restore_flags(_flags);

	return 0;
}

/********************************* vme_free_irq *****************************
 *
 *  Description: Remove handler for VME interrupt
 *	
 *	
 *---------------------------------------------------------------------------
 *  Input......: vme_irq		VME irq number to be removed
 *				 dev_id			must be the same ptr as for vme_request_irq
 *  Output.....: -
 *  Globals....:
 ****************************************************************************/
void vme_free_irq(unsigned int vme_irq, void *dev_id)
{
	struct vme_irq_entry *irqent, *prev;
	unsigned long _flags;

	if( vme_irq >= 256 ) 
		return;

	irqent = G_irqTbl[vme_irq];
	prev = NULL;

	while( irqent ){
	
		if( irqent->dev_id == dev_id )
			break;
		prev = irqent;
		irqent = prev->next;
	}
	if( irqent == NULL ){
		printk(KERN_WARNING "vme_free_irq: bad dev_id\n" );
		return;
	}

	/* remove entry from irq table */
	__save_flags(_flags);
	cli();
	
	if( prev )
		prev->next = irqent->next;
	else
		G_irqTbl[vme_irq] = irqent->next;

	__restore_flags(_flags);

	kfree( irqent );
}

/********************************* vme_ilevel_control ************************
 *
 *  Description: Enable or disable a VMEbus interrupt level
 *			   
 *			   
 *---------------------------------------------------------------------------
 *  Input......: level		VME bus irq level to modify (1..7)
 *				 enable		1=enable 0=disable
 *  Output.....: returns:	Linux error number (negative)
 *  Globals....: -
 ****************************************************************************/
int vme_ilevel_control( int level, int enable )
{
	unsigned long _flags;

	if( level < 1 || level > 7 )
		return -EINVAL;

	__save_flags(_flags);
	cli();

	if( enable ){
		VME_CTRL_SETMASK( MENPCI2VME_IMASK, 1<<level );
	}
	else {
		VME_CTRL_CLRMASK( MENPCI2VME_IMASK, 1<<level );
	}

	__restore_flags(_flags);

	return 0;
}


/* 
 * PCI Vendor/Device ID table.
 * Driver will handle all devices that have these codes
 */
static struct pci_device_id menpci2vme_pci_tbl[] __devinitdata = {
	{ MENPCI2VME_VEN_ID, MENPCI2VME_DEV_ID,
		PCI_ANY_ID, PCI_ANY_ID, },
	{ 0,}
};
MODULE_DEVICE_TABLE(pci, menpci2vme_pci_tbl);
	
static struct pci_driver menpci2vme_driver = {
	name:		"menpci2vme",
	id_table:	menpci2vme_pci_tbl,
	probe:		menpci2vme_init_one,
	remove:		__devexit_p(menpci2vme_remove_one),
};

static int __init menpci2vme_init_module(void)
{
	return pci_module_init(&menpci2vme_driver);
}

static void __exit menpci2vme_cleanup_module(void)
{
	pci_unregister_driver(&menpci2vme_driver);
}

module_init(menpci2vme_init_module);
module_exit(menpci2vme_cleanup_module);


