/***************************************************************************
                                   main.c
                             -------------------
    project:  USB device for Philips PDIUSBD12D

    begin                : Wed Dec 5 2001
    copyright            : (C) 2001 by HBM GmbH, Siegfried Wessler
    email                : siegfried.wessler@hbm.com

    latest official version:  Sep 01, 2003

    updates:  Sep 01, 2003 Wessler: Blocking read


    Source for the CP22 USB device driver.
    Module to be loaded within boot up sequence of MGCplus-CP22.
    Device name is "usbdev" as defined in "usbdev.h".

    This source is fully functional for CP22. CP42 used to have a PDIUSBD12D on board
    in very early development days. With compiler switch CP42 this source can be used
    for this early CP42. Temporarly CP42 uses a different USB chip, this source is not
    useful for.

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/*
    Just some information in advance...

    A word about types. I mislike constructs like 'unsigned short int' when we are talking about
    a 16 bit type. To run conform to HBM "standards" I only use defines that can easily be ported
    and tell the reader immediately what size this type is. So take a look at 'types.h'.
    Exeptions are made for Linux internal parameters, so these types will be adopted to newer
    kernel versions automatically.

    I don't know why some "important" speakers tell c programmers to use short names for variables etc.
    I won't bow, so whenever I didn't ported code directly into this module or didn't want to change
    names because of transparency to other HBM projects, I still tried to use simple names that should
    tell the reader what that variable stands for.


    Some more hints to work with this project:

    - You will find a define in global.h setting "DEBUG_ON". Clear the Remarks and you will get some
      more helpful "printk"s.
    - To set the hardware address or irq number look for "pdi_d12.h".
    - To change major number see "usbdev.h".


  ********************************
  Use of undocumented command in pdi_d12.c:d12_ReadMainEndpoint:

  We noticed that if you receive more than 64 bytes the second interrupt to read the second buffer of
  the double buffer is missing. So I use an undocumented command I got from the philips example code.
  Now I check whether there are data in both buffers available and get them all during one interrupt.


  ********************************
  INSMOD:

  How to insmod the module:
    Module parameter is "force_id", so do call this module like this:
      # insmod usbdev.o force_id="CP22p12345"

    or without parameter, if you'd like to use the default id:
      # insmod usbdev.o

    Hint: calling:
      # insmod usbdev.o force_id=""
    will set the id to its default (CP22p00000) either.

  Changes 10.03.2004:
    USB chip will not be initialized until a serial number has been told by application
      either by "force_id=..." or "usb_ioctl". If force_id="" (empty string) is used,
      USB chip initilization will not start either!


  ********************************
  IOCTL:

  See function "usb_ioctl" later on for information on how to use the IOCTL.

*/


#ifndef MODULE
  #define MODULE
#endif

#include <linux/module.h>
#include <asm/io.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>     // check_region()...
#include <asm/uaccess.h>      // copy_from_user()...
#include <asm/dma.h>          // request_dma, free_dma
#include <asm/irq.h>
#include <linux/sched.h>      // cli()
#include <linux/proc_fs.h>    // create_proc_read_entry
#include <linux/poll.h>       // usb_poll

#include "types.h"
#include "global.h"
#include "pdi_d12.h"
#include "usbdev.h"



UINT8   inBuffer[inBufMax];       // Buffer for data received from the USB
UINT8   outBuffer[outBufMax];     // Buffer for data to be transmitted by USB

UINT8   *pWriteAppl;
UINT8   *pSendUSB;
UINT8   *pReadAppl;
UINT8   *pGetUSB;

boolean interface_open;           // Flag to tell driver that open occured

static int major = DEV_MAJOR;

INT8  *force_id = USB_SerialNo_Str;

#ifdef DEBUG_ON
  INT32 disint = 0;
#endif

wait_queue_head_t read_waitqueue;  // waitqueue for blocking user read call operation


struct file_operations fops =
{
  ioctl:    usb_ioctl,
  open:     usb_open,
  release:  usb_close,
  read:     usb_read,
  write:    usb_write,
  poll:     usb_poll

};


DECLARE_TASKLET(tasklet_after_d12_interrupt, after_d12_interrupt, 0);

// For module documentation:
MODULE_AUTHOR(MODULE_INFO_AUTHOR);
MODULE_DESCRIPTION(MODULE_INFO_DESC);

// Module parameters. See comment at the top of this file:
MODULE_PARM (force_id, "s");



void usb_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{

  DISABLE_INTERRUPT;

  DPRINT("i");

  tasklet_schedule(&tasklet_after_d12_interrupt);     // after all interrupts are cleared call "after_d12_interrupt()"

}



INT32 init_module(void)
{
INT32 i;

  printk("%s: driver date: %s, compiled for %s\n", MODULE_NAME, DRIVER_DATE, DRIVER_TARGET);
  printk("%s: vendor-id=0x%04X, product-id=0x%04X, revision-id=0x%04X, spec-id=0x%04X\n", MODULE_NAME, USB_VENDOR_ID, USB_PRODUCT_ID, USB_REVISION_ID, USB_SPEC_ID);

  // Set up owner pointers:
  SET_MODULE_OWNER(&fops);

  i=0;
  while ((*(force_id) != 0) && (i<10))
    USB_SerialNo_Str[i++] = (*(force_id++));
  USB_SerialNo_Str[i] = 0;

  // register as character device:
  if ( (register_chrdev(major, MODULE_NAME, &fops)) != 0 )
  {
    printk("%s: register_chrdev failed.\n", MODULE_NAME);
    return -1;
  }
  else
  {
    DPRINT("%s: register_chrdev done.\n", MODULE_NAME);
  }

  i = check_region(portbase_addr, portbase_len);

  if (i == 0)
  {
    request_region(portbase_addr, portbase_len, MODULE_NAME);
    DPRINT("%s: request_region done.\n", MODULE_NAME);
  }
  else
  {
    printk("%s: check_region (%0Xh, %0Xh, %s) not successful. return: %i\n", MODULE_NAME, portbase_addr, portbase_len, MODULE_NAME, (INT16) i);
    return -1;
  }

  // register interrupt:
  if ((request_irq(USB_IRQ_NO, usb_interrupt, SA_INTERRUPT, MODULE_NAME, NULL)) != 0 )
  {
    printk("%s: request_irq for IRQ %d failed.\n", MODULE_NAME, USB_IRQ_NO);
    release_region(portbase_addr, portbase_len);
    return -1;
  }
  else
  {
    printk("%s: IRQ %d, hardware addr=0x%04X. Needed address space: %d byte.\n", MODULE_NAME, USB_IRQ_NO, portbase_addr, (1 << (PDIADDR +1)) );
  }

  create_proc_read_entry("bus/usbdev", 0, NULL, usbdev_read_procmem, NULL);

  pWriteAppl = outBuffer;
  pSendUSB = outBuffer;
  pReadAppl = inBuffer;
  pGetUSB = inBuffer;

  interface_open = FALSE;
  flag_suspend = FALSE;
  flag_bus_reset = FALSE;

  init_waitqueue_head(&read_waitqueue);

  if (USB_SerialNo_Str[0])
  { // USB_SerialNo_Str not empty = valid:
    d12_init(TRUE);   // initializes the USB chip PDIUSBD12D completely
  }

  return 0;
} //  init_module()



void cleanup_module(void)
{

  free_irq(USB_IRQ_NO, NULL);

  release_region(portbase_addr, portbase_len);

  if (unregister_chrdev(major, MODULE_NAME) != 0)
    printk("%s: Unload of driver failed.\n", MODULE_NAME);
  else
    DPRINT("%s: Driver unloaded.\n", MODULE_NAME);

} //  cleanup_module


int usb_open(struct inode *inodp, struct file *filp)
{
  DPRINT("%s: open methode called\n", MODULE_NAME);

  MOD_INC_USE_COUNT;

  interface_open = TRUE;

  pWriteAppl = outBuffer;
  pSendUSB = outBuffer;
  pReadAppl = inBuffer;
  pGetUSB = inBuffer;

  init_waitqueue_head(&read_waitqueue);

  return 0;
}


int usb_close(struct inode *inodp, struct file *filp)
{
  DPRINT("%s: close methode called\n", MODULE_NAME);

  wake_up_interruptible(&read_waitqueue);  // return outstanding read calls

  interface_open = FALSE;

  MOD_DEC_USE_COUNT;

  return 0;
} //  usb_close



ssize_t usb_write(struct file *filp, const char *userBuffer, size_t userCount, loff_t *loff)
{
// "usb_write" copies the data bytes that are to be transmitted by USB into "outBuffer".
// There are no semaphores, so be aware that you call it only once!
//
// Enter values:  file          File handle to USB driver, not accessed by usb_write.
//                *userBuffer   Pointer to user buffer with data to be transmitted.
//                userCount     Number of data to be transmitted
//                loff          not used by usb_write
//
// Return value:  Number of data bytes that could be handled to driver internal buffer "outBuffer" and probably will be transmitted next.
//                  If the return value is less than userCount, the user has to take care that the data bytes left have to be repeated
//                  the next time "usb_write" is called.

INT32   toCopy;
INT32   bytes_written;
INT32   free_right, free_left;
UINT8   *pWrite, *pSend;


  if ( (interface_open == FALSE) || flag_suspend || flag_bus_reset )
    return -EPIPE;                  // Broken Pipe

  if (userCount == 0)
    return 0;

  pWrite = pWriteAppl;
  pSend = pSendUSB;

  if (pWrite >= outBuffer +outBufMax)
  {
    if (pSend > outBuffer +1)
    { // pWriteAppl is at the right boundary and pSendUSB is not at the very first two bytes
      pWrite = outBuffer;
    }
    else
      return 0;
  }

  if (pSend <= pWrite)
  { // pWriteAppl is right of pSendUSB. if pSendUSB==pWriteAppl all data were already transmitted:
    free_right = (outBuffer -pWrite) +outBufMax;

    if (userCount <= free_right)
    {
      if (copy_from_user(pWrite, userBuffer, userCount) != 0)
        return 0;
      else
      {
        pWrite += userCount;
        bytes_written = userCount;
      }
    }
    else
    { // userBuffer doesn't fit into right part of outBuffer.
      // get first part of userBuffer:
      if (copy_from_user(pWrite, userBuffer, free_right) != 0)
        return 0;
      else
      { // now see how much of the second part of userBuffer can be read:
        bytes_written = free_right;

        if (pSend <= outBuffer +1)
        { // if pSend is at the very left boundary, there is no space available
          pWrite += free_right;
        }
        else
        {
          pWrite = outBuffer;   // set pWrite to left boundary
          free_left = (pSend -outBuffer) -1;  // always keep one byte difference!
          toCopy = userCount -free_right;

          if (free_left -toCopy >= 0)
          { // if at left boundary is enough space, get all:
            if (copy_from_user(pWrite, userBuffer +free_right, toCopy) != 0)
              return 0;
            else
            {
              pWrite += toCopy;
              bytes_written += toCopy;
            }
          }
          else
          { // pSend is too close, don't get them all:
            if (copy_from_user(pWrite, userBuffer +free_right, free_left) != 0)
              return 0;
            else
            {
              pWrite += free_left;
              bytes_written += free_left;
            }
          }
        }
      }
    }
  }
  else
  { // when usb_write was called pWriteAppl was left of pSendUSB:
    free_right = (pSend -pWrite) -1;  // always keep one byte difference!

    if (free_right == 0)
      return 0;
    else
    {
      if (free_right > userCount)
        free_right = userCount;

      if (copy_from_user(pWrite, userBuffer, free_right) != 0)
        return 0;
      else
      {
        pWrite += free_right;
        bytes_written = free_right;
      }
    }
  }

  pWriteAppl = pWrite;

  return  bytes_written;

} //  usb_write



UINT8   ReadBuffer[inBufMax];

UINT32   ApplReadBulkBuffer(char *userBuffer, INT32 userCount)
{
INT32  bytes_read;
UINT8   *pRead, *pGet;
UINT8   *pBuffer;
INT32   toCopy;

  pRead = pReadAppl;
  pGet = pGetUSB;

  if (pRead == pGet)
  { // no data to return to application
    return 0;
  }


  if (pRead >= inBuffer +inBufMax)
  { // pReadAppl at the very right boundary:
    if (pGet > inBuffer)
    {
      pRead = inBuffer;
    }
    else
      return 0;
  }

  if (pRead < pGet)
  {
    toCopy = pGet -pRead;

    if (toCopy > userCount)
      toCopy = userCount;

    if (copy_to_user(userBuffer, pRead, toCopy) != 0)
      return 0;
    else
      pRead += toCopy;

    bytes_read = toCopy;
  }
  else
  { // pRead > pGet
    toCopy = inBufMax -(pRead -inBuffer);

    if (toCopy > userCount)
    { // user doesn't want that much data; don't do a wrap:
      toCopy = userCount;
      if (copy_to_user(userBuffer, pRead, toCopy) != 0)
        return 0;
      else
        pRead += toCopy;

      bytes_read = toCopy;
    }
    else
    { // okay, user can handle this so far, so get everything right of pRead:
      pBuffer = &ReadBuffer[0];
      memcpy(pBuffer, pRead, toCopy);

      pRead += toCopy;
      pBuffer += toCopy;
      bytes_read = toCopy;

      if (pGet > inBuffer)
      { // there is some data at the very left of inBuffer valid
        toCopy = pGet -inBuffer;

        if (bytes_read +toCopy > userCount)
        { // user can't handle everything, only add as much as possible to the buffer:
          toCopy = userCount -bytes_read;
        }

        memcpy(pBuffer, inBuffer, toCopy);
        pRead = inBuffer +toCopy;
        bytes_read += toCopy;
      }

      if (copy_to_user(userBuffer, &ReadBuffer[0], bytes_read) != 0)
        return 0;
    }
  }

  pReadAppl = pRead;

  return bytes_read;
} // ApplReadBulkBuffer



ssize_t usb_read(struct file *filp, char *userBuffer, size_t userCount, loff_t *loff)
{
// "usb_read" copies the data bytes that were received by USB into the user buffer "*userBuffer".
// When no data is to be copied, it runs non blocking.
// When the user buffer can't handle that much data as already received by USB, the user buffer will be
//    filled with "count" bytes and the remaining data in the internal driver buffer will be copied to
//    offset 0 and transmitted the next time "usb_read" is called.
// There are no semaphores, so be aware that you call it only once!
//
// Important!: If the ring buffer inBuffer is running full, all data transmitted by USB from USB host
//    to CP22 will be thrown away. So keep in mind that the CP application has to call usb_read fast
//    enough, and keep a checksum over the data from USB host to CP22, e.g. when running firmware download!
//
// Enter values:  file          File handle to USB driver, not accessed by "usb_write".
//                *userBuffer   Pointer to user buffer with data that have been received.
//                userCount     Maximum number of data bytes userBuffer can take.
//                loff          Not used by "usb_write".
//
// Return value:  Number of data bytes that were copied to user "userBuffer". There still might be some data
//                  in the queue, so user: keep polling "usb_read".
//                If the device has not been previously opened or are no data to be transmitted to user,
//                  "usb_read" simply will return 0.

size_t ret_value;

  if ((interface_open == FALSE) || (userCount == 0) || flag_suspend || flag_bus_reset)
    return 0;

  if (filp->f_flags & O_NONBLOCK)
  { // read call is non-blocking; immediately try to get some data:
    ret_value = ApplReadBulkBuffer(userBuffer, userCount);
  }
  else
  { // read call is blocking:

    if (pReadAppl == pGetUSB)
    { // nothing to read, wait for some data to arrive:
      wait_event_interruptible(read_waitqueue, (pReadAppl != pGetUSB));

      // we've been awaken:
      if ( !(interface_open))
      { // interface is no longer open:
        ret_value = -1;
      }
      else
      { // interface is still open; go and get some data:
        ret_value = ApplReadBulkBuffer(userBuffer, userCount);
      }
    }
    else
    { // some data available:
      ret_value = ApplReadBulkBuffer(userBuffer, userCount);
    }

  }

  return ret_value;

} //  usb_read




int usb_ioctl(struct inode *inodep, struct file *filep, unsigned int cmd, unsigned long arg)
{
INT8  Addr[10];
UINT8 i;

  switch(cmd)
  {
    case USB_IOC_SET_ADDR:

    // To set HBM specific high level address set "arg" with pointer to character string.
    //   This string has to be exactly 10 characters long, followed by a zero byte for termiation.
    //   Usual settings will be like: "CP22p00001\0".
      copy_from_user(Addr, (INT8*) (arg), 11);
      DPRINT("%s: ioctl-cmd=%ld, *(arg)=%s\n",MODULE_NAME, cmd, Addr);

      if (Addr[10] == 0)
        for (i=0; i<11; i++)
          USB_SerialNo_Str[i] = Addr[i];
      else
        return -1;

      d12_restart();    // Restart USB connection with new Address
      // If you comment this single line, you will also get a running driver. The difference will be
      // that the string id of the usb device will be changed, but the pipe to the host pc will still
      // exist. This might be an advantange, if you don't like Win2000 pop up screens. But to achieve
      // this function right, the windows/host pc will have to poll the whole information out of this
      // usb device driver whenever a descriptor is needed. So far, some information is stored on the
      // windows/host pc. That's why I restart the usb chip, so to the windows/host pc this looks
      // like an unplugging of the usb device and all descriptors are exchanged again.

      break;


    default:
      DPRINT("%s: ioctl-cmd=%ld, arg=%ld\n", MODULE_NAME, cmd, arg);
      return -1;
  }
  return 0;
}



unsigned int usb_poll(struct file *filp, poll_table *wait)
{
unsigned int  mask;


  DPRINT("%s: poll methode called\n", MODULE_NAME);

  mask = 0;

  if (interface_open)
  {
    poll_wait(filp, &read_waitqueue, wait);
    if (pGetUSB != pReadAppl)
      mask |= (POLLIN | POLLRDNORM);
    if ((pSendUSB -pWriteAppl) != 2)
      mask |= (POLLOUT | POLLWRNORM);
  }

  return mask;

}   // usb_poll








int usbdev_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
int len;

  DPRINT("%s: read_procmem, count=%i\n", MODULE_NAME, count);

  len = 0;

  if (count > 200)
  {
    len = sprintf(buf,"version: %s, date: %s\n", DRIVER_VERSION, DRIVER_DATE);  // 19 +20
    len += sprintf(buf+len, "USB-SerialNo: [%s]\n", USB_SerialNo_Str);                 // 25
    len += sprintf(buf+len, "USB device side driver for CP22\n");                       // 35
    len += sprintf(buf+len, "by HBM GmbH, Dept. T-V, Siegfried Wessler\n");         // 44
  }

  *eof = 1;     // no more data to come

  return len;

}
