/***************************************************************************
                                  pdi_d12.c
                             -------------------
    project:  USB device for Philips PDIUSBD12D

    begin                : Wed Dec 5 2001
    copyright            : (C) 2001 by HBM GmbH, Siegfried Wessler
    email                : siegfried.wessler@hbm.com

    latest official version:  May 09, 2003

    updates:  May 08, 2003 Wessler: changes to "rep_outs" using "outsl" instead
                                    d12_init uses parameter

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include  <asm/io.h>

#include  "types.h"
#include  "global.h"
#include  "pdi_d12.h"
#include  "usbdev.h"
#include  <asm/irq.h>
#include  <linux/sched.h>   // cli()


// *******************************************

#define ZEROPACKET_DELAY  2;
boolean fullblock_written = FALSE;
UINT8   zeropacket_delay_count = 0;

// *******************************************


UINT8 read_usbd(void)
{ // Read only one data byte and return it
    
  return  inportb(D12_DATA);
}


void  write_usbd(UINT8 data)
{ // Simply write 1 data
  outportb(D12_DATA, data);
}


UINT8 read_usbc1(UINT8 addr)
{ // write command with address 'addr' and return 1 data byte
  outportb(D12_COMMAND, addr);

  return (inportb(D12_DATA));
}


UINT16 read_usbc2(UINT8 addr)
{ // write command with address 'addr' and receive 2 data bytes

  UINT8   data1;
  UINT16  data2;
  
  outportb(D12_COMMAND, addr);
  data1 = inportb(D12_DATA);
  data2 = inportb(D12_DATA);
  
  data2 = (data2 << 8) | data1;
  
  return data2;
}


void write_usbc0(UINT8 addr)
{ // write command with address 'addr'
  outportb(D12_COMMAND, addr);
}


void write_usbc1(UINT8 addr, UINT8 data)
{ // write command with address 'addr' and write 1 data byte
  outportb(D12_COMMAND, addr);
  outportb(D12_DATA, data);
}


void write_usbc2(UINT8 addr, UINT8 data1, UINT8 data2)
{ // write command with address 'addr' and write 2 data byte
  outportb(D12_COMMAND, addr);
  outportb(D12_DATA, data1);
  outportb(D12_DATA, data2);
}




// ########################################



void  d12_init(boolean reset)
{
INT16 i;

  DISABLE_INTERRUPT;

  if (reset)
  {
    printk("%s: Initialize USB device. USB-SerialNo=[%s]\n", MODULE_NAME, USB_SerialNo_Str);

    DPRINT("%s: Reading Interrupt edge/level register: I/O port 0x4D1=0x%X, 0x4D0=0x%X\n", MODULE_NAME, inb(0x4d1), inb(0x4d0));
    DPRINT("%s: Setting USB interrupt to level sensitivity.\n", MODULE_NAME);

    SET_INTERRUPT_LEVEL;      // Set processors interrupt sensitivity for use of USB irq to level
    DPRINT("%s: Reading Interrupt edge/level register: I/O port 0x4D1=0x%X, 0x4D0=0x%X\n", MODULE_NAME, inb(0x4d1), inb(0x4d0));
  }
  else
  {
    printk("%s: Reinitialize USB device. USB-SerialNo=%s\n", MODULE_NAME, USB_SerialNo_Str);
  }


  // initialize clock generator:
  // Prior to this clock output will be 4 MHz, after it will be (48 MHz / CLOCKDIV).
  i = 12;
  do
  {
    write_usbc2(D12_SET_MODE, D12_CLOCK_RUNNING | D12_NO_LAZYCLOCK, D12_SET_TO_ONE +i);
  } while ((--i) != (CLOCKDIV -1));

  delayc(1000);

  write_usbc1(D12_SET_DMA, D12_ENDPX4_INT_ENABLE | D12_ENDPX5_INT_ENABLE);   // no DMA, but interrupts
  write_usbc2(D12_SET_MODE, D12_SOFTCONNECT | D12_INTERRUPT_MODE | D12_CLOCK_RUNNING | D12_NO_LAZYCLOCK, D12_SET_TO_ONE +CLOCKDIV -1);   // softconnect, all interrupts

  d12_ClearEndpX(5);
  d12_ClearEndpX(4);

  delayc(1000);

  ENABLE_INTERRUPT;

}   // d12_init()




UINT8 d12_ReadEndpoint(UINT8 endpx, UINT8 * buf, UINT32 len)
{
UINT8 i, j;

  DPRINT("- ReadEP%d: ", endpx);

  if ((read_usbc1(endpx) & D12_FULLEMPTY) == 0)
  {
    return 0;
  }

  j = read_usbc1(D12_READ_BUFFER);    // Dummy read: reserved
  j = read_usbd();                    // next byte: number of data bytes
  
  if (j > len)
    j = len;
    
  for (i=0; i<j; i++)
    *(buf +i) = read_usbd();
    
  write_usbc0(D12_CLEAR_BUFFER);
  
  DPRINT("%d\n", j);
  
  return j;
}   // d12_ReadEndpoint



UINT16 d12_ReadMainEndpoint(UINT8 * buf, UINT32 len)
{
UINT16 i = 0;
UINT8 doublebuf = 1;

UINT16 k = 0;
UINT8 *bufp;
  
  DPRINT("- ReadMainEPx4");

  // not documented, but highly recommended by me. See example application D12CI.c
  write_usbc0(0x84);  
  if ((read_usbd() & 0x60) == 0x60)
  {
    doublebuf = 2;
    DPRINT(" ** DOUBLE BUFFERING! **");
  }

  while (doublebuf)
  {
    write_usbc0(D12_SELECT_ENDPX4);

    if ((read_usbc1(D12_SELECT_ENDPX4) & D12_FULLEMPTY) == 0)
      break;


    i = read_usbc1(D12_READ_BUFFER);    // Dummy read: reserved
    i = read_usbd();                    // next byte: number of data bytes
  
    if (i+k > len)
      i = len -k;
    
    bufp = buf+k;
    insb(D12_DATA, bufp, i);
    k += i;
    

    write_usbc0(D12_CLEAR_BUFFER);
  
    doublebuf--;
  }
  
  DPRINT("%d\n", k);

  return k;
}   // d12_ReadMainEndpoint



UINT8 d12_WriteEndpoint(UINT8 endpx, UINT8 * buf, UINT32 len)
{
UINT8 i;

  read_usbc1(D12_SELECT_ENDPX +endpx);  

  write_usbc2(D12_WRITE_BUFFER, 0, len);

  if (len > 0)
  {
    DPRINT("I%02X -  WriteEP%d: control_state=%d, Daten (%d) ", int_status, endpx, control_state, len); 
    for(i=0; i<len; i++)
      write_usbd( *(buf +i) );
  }

  write_usbc0(D12_VALIDATE_BUFFER);

  DPRINT("%d\n", len);

  return len;
} //  d12_WriteEndpoint



UINT8 d12_WriteMainEndpoint(UINT8 * buf, UINT32 len)
{
UINT8 sent;

  if (len >= EP2_TX_FIFO_SIZE)
  {   // send full block:
    DPRINT("WriteMainEP: control_state=%d, Daten (%d) - full block\n", control_state, len);
    sent = EP2_TX_FIFO_SIZE;
    read_usbc1(D12_SELECT_ENDPX5);
    write_usbc2(D12_WRITE_BUFFER, 0, sent);

#if PDIADDR==0
// only byte transfers possible:
    outsb(D12_DATA, buf, sent);
#else
  #if PDIADDR==1
  // byte and word transfers possible:
    if ((sent & 0x01) == 0)
      outsw(D12_DATA, buf, sent);   // word transfer
    else
      outsb(D12_DATA, buf, sent);   // byte transfer
  #else
    // take long transfer, if possible, otherwise word or byte transfer:
    if ((sent & 0x03) == 0)
      outsl(D12_DATA, buf, sent >> 2);     // long transfer
    else
      if ((sent & 0x01) == 0)
        outsw(D12_DATA, buf, sent >> 1);   // word transfer
      else
        outsb(D12_DATA, buf, sent);   // byte transfer
  #endif
#endif

    write_usbc0(D12_VALIDATE_BUFFER);

    fullblock_written = TRUE;
    zeropacket_delay_count = ZEROPACKET_DELAY;
  }
  else
  {   // block < EP2_TX_FIFO_SIZE:
  
    sent = len;

    if (len > 0)
    {   // some data to send
      DPRINT("WriteMainEP: control_state=%d, Daten (%ld) - some data\n", control_state, len);
      read_usbc1(D12_SELECT_ENDPX5);
      write_usbc2(D12_WRITE_BUFFER, 0, sent);

#if PDIADDR==0                 // only byte transfers possible
      outsb(D12_DATA, buf, sent);
#else
  #if PDIADDR==1               // byte and word transfers possible
      if ((sent & 0x01) == 0)
        outsw(D12_DATA, buf, sent);   // word transfer
      else
        outsb(D12_DATA, buf, sent);   // byte transfer
  #else
      if ((sent & 0x03) == 0)
        outsl(D12_DATA, buf, sent >> 2);     // long transfer
      else
        if ((sent & 0x01) == 0)
          outsw(D12_DATA, buf, sent >> 1);   // word transfer
        else
          outsb(D12_DATA, buf, sent);   // byte transfer
  #endif
#endif

      write_usbc0(D12_VALIDATE_BUFFER);

      fullblock_written = FALSE;
      zeropacket_delay_count = ZEROPACKET_DELAY;
    }
    else
    {   // nothing to send
      if ( (fullblock_written) || ((--zeropacket_delay_count) == 0) )
      {   // send a zero length packet only every ZEROPACKET_DELAY times:
        read_usbc1(D12_SELECT_ENDPX5);
        write_usbc2(D12_WRITE_BUFFER, 0, 0);
        write_usbc0(D12_VALIDATE_BUFFER);
      
        fullblock_written = FALSE;

        zeropacket_delay_count = ZEROPACKET_DELAY;
      }
    }

  }

  return sent;

} //  d12_WriteMainEndpoint



void  d12_AcknowledgeEndpoint(UINT8 endpx)
{
  
  DPRINT("  AcknowledgeEP%d\n", endpx);
  
  write_usbc0(endpx);
  write_usbc0(D12_ACK_SETUP);
  
  if (endpx == 0)
    write_usbc0(D12_CLEAR_BUFFER);
    
} //  d12_AcknowledgeEndpoint



void  d12_SetEndpointStatus(UINT8 endpx, UINT8 stalled)
{
  
  DPRINT("  SetEPStatus%d: stalled=%d\n", endpx, stalled);
  
  write_usbc1(D12_SELECT_ENDPX +endpx, stalled);
  
} //  d12_SetEndpointStatus



void  d12_ClearEndpX(UINT8 EPIndex)
{

    write_usbc0(D12_SELECT_ENDPX0+EPIndex);
    write_usbc0(D12_CLEAR_BUFFER);
}

