/***************************************************************************
                                   main.c
                              -------------------
    project:  USB device for Philips ISP1181A

    begin                : Wed Sep 4 2002
    copyright            : (C) 2002 by HBM GmbH, Siegfried Wessler
    email                : siegfried.wessler@hbm.com

    Source for the CP42 USB device driver.
    Module to be loaded within boot up sequence of MGCplus-CP42.
    Device name is "usbdev" as defined in "usbdev.h".

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/*
    Just some information in advance...

    A word about types. I mislike constructs like 'unsigned short int' when we are talking about
    a 16 bit type. To run conform to HBM "standards" I only use defines that can easily be ported
    and tell the reader immediately what size this type is. So take a look at 'types.h'.
    Exeptions are made for Linux internal parameters, so these types will be adopted to newer
    kernel versions automatically.

    I don't know why some "important" speakers tell c programmers to use short names for variables etc.
    I won't bow, so I still tried to use simple names that should tell the reader what that variable
    stands for. Exeptions I made when I directly ported code or didn't want to change names because
    of transparency to other HBM projects.


    Some more hints to work with this project:

    - You will find a define in global.h setting "DEBUG_ON". Clear the Remarks and you will get some
      more helpful "printk"s.
    - To set the hardware address or irq number look for "isp1181.h".
    - To change major number see "usbdev.h".
    - Whenever you see "IN" this referenes the direction from CP42 to the usb host, whereas "OUT" gives
      the direction from the usb host to the CP42.

    Some words about double buffering:
    To save fifo space for isochronous endpoints, double buffering of bulk endpoints is switched off,
    if AlternateSetting 1 is activated. Isochronous endpoints always use double buffering.
    When AlternateSetting 0 is active, no isochronous endpoints are available (due to USB specification)
    and double buffering for bulk endpoints is active.


  ********************************
  INSMOD:

  How to insmod the module:
    Module parameter is "force_id", so do call this module like this:
      # insmod usbdev.o force_id="CP42p12345"

    or without parameter, if you'd like to use the default id:
      # insmod usbdev.o

    Hint: calling:
      # insmod usbdev.o force_id=""
    will set the id to its default (CP42p00000) either.

  Changes 10.03.2004:
    USB chip will not be initialized until a serial number has been told by application
      either by "force_id=..." or "usb_ioctl". If force_id="" (empty string) is used,
      USB chip initilization will not start either!


  ********************************
  IOCTL:

  See function "usb_ioctl" later on for information on how to use the IOCTL.

*/


#ifndef MODULE
  #define MODULE
#endif

#include <linux/module.h>
#include <asm/io.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>    // check_region()...
#include <asm/uaccess.h> // copy_from_user()...
#include <asm/semaphore.h>
#include <linux/poll.h>   // poll/select
#include <linux/fcntl.h>
#include <linux/proc_fs.h>  // create_proc_read_entry

#include "types.h"
#include "global.h"
#include "isp1181.h"
#include "descript.h"
#include "usbdev.h"


tDevice main_dev;
tDevice iso1_dev;
tDevice iso2_dev;
tDevice iso3_dev;

tBulkBuffer MAINBuffer;

tIsoBuffer  IsoBufferA[3];
tIsoBuffer  IsoBufferB[3];
tIsoBuffer  *WrIsoBuffer[3];
tIsoBuffer  *RdIsoBuffer[3];

UINT8   AlternateSetting;         // Number of Alternate Interface Setting, set by usb request "Set Interface"
boolean EnableBulkDoubleBuffer;    // TRUE, if BULK IN/OUT is double buffered; switched to FALSE, while AlternateSetting != 0

boolean EnableSOFInt;



static int major = DEV_MAJOR;

INT8  *force_id = USB_SerialNo_Str;



struct file_operations fops =
{
  ioctl:    usb_ioctl,
  open:     usb_open,
  release:  usb_close,
  read:     usb_read,
  write:    usb_write,
  poll:     usb_poll

};


DECLARE_TASKLET(tasklet_after_isp_interrupt, after_isp_interrupt, 0);

// For module documentation:
MODULE_AUTHOR(MODULE_INFO_AUTHOR);
MODULE_DESCRIPTION(MODULE_INFO_DESC);
MODULE_LICENSE(MODULE_INFO_LICENSE);
MODULE_SUPPORTED_DEVICE(MODULE_INFO_SUPPORT);

// Module parameters. See comment at the top of this file:
MODULE_PARM (force_id, "s");



void usb_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{

  DISABLE_INTERRUPT;

  printd("i");

  tasklet_schedule(&tasklet_after_isp_interrupt);     // after all interrupts are cleared call "after_isp_interrupt()"

}



INT32 init_module(void)
{
INT32 i;

  DTRIGGER_ON;

  printk("%s: driver version: %s, driver date: %s, compiled for %s\n", MODULE_NAME, DRIVER_VERSION, DRIVER_DATE, DRIVER_TARGET);

  // Set up owner pointers:
  SET_MODULE_OWNER(&fops);

  i=0;
  while ((*(force_id) != 0) && (i < SERIALNUM_LENGTH))
    USB_SerialNo_Str[i++] = *(force_id++);
  USB_SerialNo_Str[i] = 0;

  copy_serialstr_to_serialdescr();

  // register as character device:
  if ( (register_chrdev(major, MODULE_NAME, &fops)) != 0 )
  {
    printk("%s: register_chrdev failed.\n", MODULE_NAME);
    return -1;
  }
  else
  {
    DPRINT("%s: register_chrdev done.\n", MODULE_NAME);
  }

  i = check_region(portbase_addr, portbase_len);

  if (i == 0)
  {
    request_region(portbase_addr, portbase_len, MODULE_NAME);
    DPRINT("%s: request_region done.\n", MODULE_NAME);
  }
  else
  {
    printk("%s: check_region (%0Xh, %0Xh, %s) not successful. return: %i\n", MODULE_NAME, portbase_addr, portbase_len, MODULE_NAME, (INT16) i);
    return -1;
  }

  // register interrupt:
  if ((request_irq(USB_IRQ_NO, usb_interrupt, SA_INTERRUPT, MODULE_NAME, NULL)) != 0 )
  {
    printk("%s: request_irq for IRQ %d failed.\n", MODULE_NAME, USB_IRQ_NO);
    release_region(portbase_addr, portbase_len);
    return -1;
  }
  else
  {
    printk("%s: IRQ %d, hardware addr=0x%04X installed as USB device.\n", MODULE_NAME, USB_IRQ_NO, portbase_addr);
  }

  create_proc_read_entry("bus/usbdev", 0, NULL, usbdev_read_procmem, NULL);

  main_dev.flags &= ~(FLAG_INTERFACE_OPEN);
  sema_init(&main_dev.sem, 1);

  iso1_dev.flags &= ~(FLAG_INTERFACE_OPEN);
  sema_init(&iso1_dev.sem, 1);

  iso2_dev.flags &= ~(FLAG_INTERFACE_OPEN);
  sema_init(&iso2_dev.sem, 1);

  iso3_dev.flags &= ~(FLAG_INTERFACE_OPEN);
  sema_init(&iso3_dev.sem, 1);

  spin_lock_init(&Lock_FeedBulkIN);
  spin_lock_init(&Lock_FeedIso);

  InitMAINBuffer();

  for (i=0; i<3; i++)
  {
    IsoBufferA[i].ToSend = 0;
    IsoBufferB[i].ToSend = 0;
    IsoBufferA[i].Status = 0;
    IsoBufferB[i].Status = 0;
    IsoBufferA[i].Offset = 0;
    IsoBufferB[i].Offset = 0;
    IsoBufferA[i].Index = 1;
    IsoBufferB[i].Index = 2;
    WrIsoBuffer[i] = &IsoBufferA[i];
    RdIsoBuffer[i] = &IsoBufferA[i];
  }

  flag_suspend = FALSE;
  flag_bus_reset = FALSE;

  AlternateSetting = 0;
  EnableBulkDoubleBuffer = TRUE;     // set default: Bulk double buffer enabled

  EnableSOFInt = FALSE;

  if (USB_SerialNo_Str[0])
  { // USB_SerialNo_Str not empty = valid:
    isp_init(TRUE);   // initializes the USB chip completely
  }

  return 0;
} //  init_module()



void cleanup_module(void)
{

  free_irq(USB_IRQ_NO, NULL);

  release_region(portbase_addr, portbase_len);

  if (unregister_chrdev(major, MODULE_NAME) != 0)
    printk("%s: Unload of driver failed.\n", MODULE_NAME);
  else
  {
    DPRINT("%s: Driver unloaded.\n", MODULE_NAME);
  }

} //  cleanup_module


int usb_open(struct inode *inodp, struct file *filp)
{
tDevice   *dev;
UINT16    minor_no;

  minor_no = MINOR(inodp->i_rdev);
  DPRINT("%s: open methode called, minor-no=%d\n", MODULE_NAME, minor_no);

  switch (minor_no)
  {
    case  0:  dev = &main_dev;   // Bulk MAIN_IN / MAIN_OUT
              InitMAINBuffer();
              break;
    case  1:  dev = &iso1_dev;   // Iso Pipe 1 = Sample rate group 1
              break;
    case  2:  dev = &iso2_dev;   // Iso Pipe 2 = Sample rate group 2
              break;
    case  3:  dev = &iso3_dev;   // Iso Pipe 3 = Sample rate group 3
              break;
    default:  printk("%s: open - error: minor_no=%d\n", MODULE_NAME, minor_no);
              return -ENODEV;
              break;
  }

  MOD_INC_USE_COUNT;

  dev->flags |= (FLAG_INTERFACE_OPEN);
  dev->minor_no = minor_no;
  sema_init(&dev->sem, 1);

  filp->private_data = dev;

  if (minor_no)
  {
    IsoBufferA[minor_no-1].ToSend = 0;
    IsoBufferA[minor_no-1].Status = 0;
    IsoBufferA[minor_no-1].Offset = 0;
    IsoBufferB[minor_no-1].ToSend = 0;
    IsoBufferB[minor_no-1].Status = 0;
    IsoBufferB[minor_no-1].Offset = 0;
  }

  return 0;
}   // usb_open


int usb_close(struct inode *inodp, struct file *filp)
{
tDevice   *dev;

  DPRINT("%s: close methode called, minor_no=%d\n", MODULE_NAME, MINOR(inodp->i_rdev));

  dev = filp->private_data;
  if (dev == NULL)
  {
    printk("%s: close - warning: object is NULL\n", MODULE_NAME);
    return -ENODEV;
  }

  down (&dev->sem);   // lock the device

  dev->flags &= ~(FLAG_INTERFACE_OPEN);

  if (dev->minor_no == 0)
    wake_up_interruptible(&MAINBuffer.waitqueue);  // return outstanding read calls

  MOD_DEC_USE_COUNT;

  up (&dev->sem);   // unlock the device
  return 0;
} //  usb_close




ssize_t usb_write(struct file *filp, const char *userBuffer, size_t userCount, loff_t *loff)
{
// bulk transfer:
// "usb_write" copies the data bytes that are to be transmitted by USB into "outBuffer".
//
// Enter values:  file          File structure to USB driver.
//                *userBuffer   Pointer to user buffer with data to be transmitted.
//                userCount     Number of data to be transmitted.
//                loff          Not used by usb_write
//
// Return value:  Number of data bytes that could be handled to driver internal buffer "outBuffer" and probably will be transmitted next.
//                  If the return value is less than userCount, the user has to take care that the data bytes left have to be repeated
//                  the next time "usb_write" is called.

INT32  calc_transmit;
tDevice *dev;

  dev = filp->private_data;
  if (dev == NULL)
  {
    printk("%s: write - warning: object is NULL\n", MODULE_NAME);
    return -ENODEV;
  }

  down (&dev->sem);   // lock the device

  if ( ((dev->flags & FLAG_INTERFACE_OPEN) == 0) || flag_suspend || flag_bus_reset )
  {
    up (&dev->sem);   // unlock the device
    return -EPIPE;    // Broken Pipe
  }

  if (userCount == 0)
  { // no data to be written:
    calc_transmit = 0;
  }
  else
  {

    switch (dev->minor_no)
    {
      case  0:  calc_transmit = ApplWriteBulkBuffer(&MAINBuffer, userBuffer, userCount);
                printd1("Write%d (%ld/%d):[%04X]%.10s\n", dev->minor_no, calc_transmit, userCount,*userBuffer, userBuffer);
                break;

      case  1:
      case  2:
      case  3:  calc_transmit = ApplWriteIsoBuffer(dev, userBuffer, userCount);
                printd("^%d\n", userCount);
                break;

      default:  calc_transmit = -1;
    }
  }

  up (&dev->sem);   // unlock the device
  return calc_transmit;

} //  usb_write



ssize_t usb_read(struct file *filp, char *userBuffer, size_t userCount, loff_t *loff)
{
// "usb_read" copies the data bytes that were received by USB into the user buffer "*userBuffer".
// It can be switch from blocking to non-blocking in reverse by IOCtrl.
// In non-blocking mode (default) it returns 0, if no data are available. In blocking mode it waits for
//  data received within function "GetBulkOUTFifo" and then returns.
//  When the user buffer can't handle that much data as already received by USB, the user buffer will be
//    filled with "count" bytes and the remaining data in the internal driver buffer will be transmitted
//    the next time "usb_read" is called.
//
// Enter values:  file          File structure to USB driver.
//                *userBuffer   Pointer to user buffer with data that have been received.
//                userCount     Maximum number of data bytes userBuffer can take.
//                                A value of 0 is allowed and will return 0 immediately, even if blocking
//                                mode was enabled.
//                loff          Not used by "usb_write".
//
// Return value:  Number of data bytes that were copied to user "userBuffer". There still might be some data
//                  in the queue, so user: keep polling "usb_read".
//                If the device has not been previously opened or the USB connection was interrupted by a
//                  USB bus reset or USB suspend "usb_read" simply will return 0.

size_t  ret_value;
tDevice *dev;


  dev = filp->private_data;
  if (dev == NULL)
  {
    printk("%s: read - warning: object is NULL\n", MODULE_NAME);
    return -ENODEV;
  }

  down (&dev->sem);   // lock this device

  if (( (dev->flags & FLAG_INTERFACE_OPEN) == 0) || (userCount == 0) || flag_suspend || flag_bus_reset)
  {
    ret_value = 0;
  }
  else
  {

    switch  (dev->minor_no)
    {
      case  0:  if (filp->f_flags & O_NONBLOCK)
                { // read call is non-blocking; immediately try to get some data:
                    ret_value = ApplReadBulkBuffer(&MAINBuffer, userBuffer, userCount);
                }
                else
                { // read call is blocking:

                  if (MAINBuffer.pReadAppl == MAINBuffer.pGetUSB)
                  { // nothing to read, wait for some data to arrive:
                    wait_event_interruptible(MAINBuffer.waitqueue, (MAINBuffer.pReadAppl != MAINBuffer.pGetUSB));

                    // we've been awaken:
                    if ( !(dev->flags & FLAG_INTERFACE_OPEN))
                    { // interface is no longer open:
                      ret_value = -1;
                    }
                    else
                    { // interface is still open; go and get some data:
                      ret_value = ApplReadBulkBuffer(&MAINBuffer, userBuffer, userCount);
                    }
                  }
                  else
                  { // some data available:
                    ret_value = ApplReadBulkBuffer(&MAINBuffer, userBuffer, userCount);
                  }

                }

if (ret_value)  //@@
  printd1("ApplReadBulk (%d/%d):%.4s\n", ret_value, userCount, userBuffer); //@@

                break;

      default:  ret_value = -1;
                break;
    }
  }

  up (&dev->sem);   // unlock the device
  return ret_value;

} //  usb_read




int usb_ioctl(struct inode *inodp, struct file *filp, unsigned int cmd, unsigned long arg)
{
// cmd:   Command byte, directly to be used as such.
// arg:   See "switch (cmd)" instructions below.

INT8    Addr[10];
UINT8   i;
tDevice *dev;
UINT8   minor_no;
INT16   ret = 0;
UINT16  setfifo;
boolean flag;

  dev = filp->private_data;
  if (dev == NULL)
  {
    printk("%s: ioctl - warning: object is NULL\n", MODULE_NAME);
    return -ENODEV;
  }

  down (&dev->sem);   // lock this object
  minor_no = dev->minor_no;

  DPRINT("%s: ioctl methode called, minor_no=%d, cmd=%d, arg=0x%04X %04X\n", MODULE_NAME, minor_no, cmd, (UINT16)(arg>>16), (UINT16)(arg&0xffff));

  switch (cmd)
  {
    case USB_IOC_SET_BUFSIZE:
      // arg: Bits 11..0:  PacketBufferSize
      //      Bit  12:     SOF-Framecounter, 0: disabled, 1: enabled (Bit is ignored in this version)
      //      Bits 15..13: reserved
      //      Bits 31..16: ScanPatternLength (Bits are ignored in this version)

      setfifo = ((UINT16) arg) & IOCTL_BUFSIZE_MASK;   // mask bits 11..0

      if ( (setfifo > 1023) || (minor_no > MAX_ENDPOINTS -2) || (minor_no < 1) )
      {
        printk("%s: Strange ioctl-arg received: minor-no=%d, ioctl-arg=0x%04X %04X, setfifo=%d\n", MODULE_NAME, minor_no, (UINT16)(arg>>16), (UINT16)(arg & 0xFFFF), setfifo);
        ret = -1;
      }
      else
        ret = calc_buffersizes(setfifo, minor_no);

      if (arg & IOCTL_SOF_BIT)
        EnableSOFInt = TRUE;
      else
        EnableSOFInt = FALSE;

      break;


    case USB_IOC_SET_SERNUM:
      // To set HBM specific high level address (ID-String), set "arg" with pointer to a zero
      //   terminated character string. In USB terminology this string sets individual serialnumber.
      //   This string has to be maximum 10 characters long, followed by a zero byte for termiation.
      //   Usual settings will look like this: "CP42p00001\0".
      // arg: 32-bit-Pointer to user space where argument is found. You have to use "copy_from_user".

      if (minor_no != 0)
      {
        ret = -1;
        goto exit;  // changing of serialnumber only allowed to device with minor=0, usually: "/dev/usbdev"
      }

      copy_from_user(Addr, (INT8*) (arg), 11);
      DPRINT("%s: ioctl-cmd=%d, *(arg)=%s\n", MODULE_NAME, cmd, Addr);

      if (Addr[10] == 0)
      {
        flag = FALSE;
        for (i = 0; i < SERIALNUM_LENGTH; i++)
        {
          if (USB_SerialNo_Str[i] != Addr[i])
          {
            flag = TRUE;
            USB_SerialNo_Str[i] = Addr[i];
          }
        }
        copy_serialstr_to_serialdescr();
      }
      else
      {
        ret = -1;
        goto exit;
      }

      up (&dev->sem);   // unlock the device

      if (flag)
        isp_restart();    // Restart USB connection with new ID-String.
      // If you comment this single line, you will also get a running driver. The difference will be
      //   that the string id of the usb device will be changed, but the pipe to the host pc will still
      //   exist. This might be an advantange, if you don't like Win2000 pop up screens. But to achieve
      //   this function right, the windows/host pc will have to poll the whole information out of this
      //   usb device driver whenever a descriptor is needed. So far, some information is stored on the
      //   windows/host pc. That's why I restart the usb chip, so to the windows/host pc this looks
      //   like an unplugging of the usb device and all descriptors are exchanged again.
      return 0;
      break;


    default:
      printk("%s: invalid argument: ioctl-cmd=%d, arg=0x%08lX\n", MODULE_NAME, cmd, arg);
      ret = -ENOTTY;    // POSIX likes -ENOTTY rather than -EINVAL
      goto exit;
  }

exit:
  up (&dev->sem);   // unlock the device
  return ret;

}   //  usb_ioctl



unsigned int usb_poll(struct file *filp, poll_table *wait)
{
tDevice *dev;
UINT8   minor_no;
unsigned int  mask;

  dev = filp->private_data;
  if (dev == NULL)
  {
    printk("%s: poll - warning: object is NULL\n", MODULE_NAME);
    return -ENODEV;
  }

  down (&dev->sem);   // lock this object
  minor_no = dev->minor_no;

  printd1("%s: poll methode called, minor_no=%d\n", MODULE_NAME, minor_no);

  mask = 0;
  switch  (minor_no)
  {
    case  0:  poll_wait(filp, &MAINBuffer.waitqueue, wait);
              if (MAINBuffer.pGetUSB != MAINBuffer.pReadAppl)
                mask |= (POLLIN | POLLRDNORM);
              if ((MAINBuffer.pSendUSB -MAINBuffer.pWriteAppl) != 2)
                mask |= (POLLOUT | POLLWRNORM);
              break;

    default:  printk("%s: error: select/poll only available for device 0, called device no.=%d\n", MODULE_NAME, minor_no);
              mask = 0;
              break;
  }


  up (&dev->sem);   // unlock the device
  return mask;

}   // usb_poll




int usbdev_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
int len;

  printd1("%s: read_procmem, count=%i\n", MODULE_NAME, count);

  len = 0;

  if (count > 200)
  {
    len = sprintf(buf,"version: %s, date: %s\n", DRIVER_VERSION, DRIVER_DATE);  // 19 +20
    len += sprintf(buf+len, "USB-SerialNo: [%s]\n", USB_SerialNo_Str);                 // 25
    len += sprintf(buf+len, "USB device side driver for CP42\n");                       // 35
    len += sprintf(buf+len, "by HBM GmbH, Dept. T-V, Siegfried Wessler\n");         // 44
  }

  *eof = 1;     // no more data to come

  return len;

}