/*
 * i2c-algo-8xx.c i2x driver algorithms for MPC8XX CPM
 * Copyright (c) 1999 Dan Malek (dmalek@jlc.net).
 *
 * moved into proper i2c interface; separated out platform specific 
 * parts into i2c-rpx.c
 * Brad Parker (brad@heeltoe.com)
 *
 * Hacked to work with either 8xx or 8260 cpm and renamed
 * i2c-algo-cpm.c by Murray.Jensen@cmst.csiro.au, 17-May-01
 */

/* $Id: i2c.patch,v 1.2 2002/01/29 11:49:09 pba Exp $ */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>

#if defined(CONFIG_8xx)
#include <asm/mpc8xx.h>
#include "../../arch/ppc/8xx_io/commproc.h"
#elif defined(CONFIG_8260)
#include <asm/mpc8260.h>
#include <asm/cpm_8260.h>
#endif

#include <linux/i2c.h>
#include <linux/i2c-algo-cpm.h>

#define BD_SC_NAK		((ushort)0x0004) /* NAK - did not respond */
#define BD_SC_UN		((ushort)0x0002) /* underrun */
#define BD_SC_CL		((ushort)0x0001) /* collision */
#define BD_SC_XER		(BD_SC_NAK|BD_SC_UN|BD_SC_CL) /* transmit err */

#define BD_SC_START		((ushort)0x0400)

#define CPM_CR_CLOSE_RXBD	((ushort)0x0007)

#define I2C_EN                  0x11             /* set i2mod to activate */
                                                 /* page 34-6 disable GCD, enable */
#define I2C_DIS			0x00		 /* reset value of i2c controller */

int cpm_iic_scan = 0;
int cpm_iic_debug = 0;

static void
cpm_iic_interrupt(void *dev_id)
{
	struct i2c_algo_cpm_data *cpm = (struct i2c_algo_cpm_data *)dev_id;
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif

	if (cpm_iic_debug > 1)
		printk("cpm_iic_interrupt(dev_id=%p)\n", dev_id);

	/* Clear interrupt.
	*/
	i2c->i2c_i2cer = 0xff;

	/* Get 'em going again.
	*/
	wake_up_interruptible(&cpm->waitq);
}

static void 
cpm_iic_reset_params(struct i2c_algo_cpm_data *cpm)
{
	volatile iic_t *iip = cpm->iip;

	/* Initialize the parameter ram.
	 * We need to make sure many things are initialized to zero,
	 * especially in the case of a microcode patch.
	 */
	iip->iic_rstate = 0;
	iip->iic_rdp = 0;
	iip->iic_rbptr = 0;
	iip->iic_rbc = 0;
	iip->iic_rxtmp = 0;
	iip->iic_tstate = 0;
	iip->iic_tdp = 0;
	iip->iic_tbptr = 0;
	iip->iic_tbc = 0;
	iip->iic_txtmp = 0;

	/* Set up the IIC parameters in the parameter ram.
	*/
	iip->iic_tbase = cpm->dp_tbase;
	iip->iic_rbase = cpm->dp_rbase;

#if defined(CONFIG_8xx)
	iip->iic_tfcr = SMC_EB;
	iip->iic_rfcr = SMC_EB;
#elif defined(CONFIG_8260)
	iip->iic_tfcr = CPMFCR_EB;
	iip->iic_rfcr = CPMFCR_EB;
#endif

	/* Set maximum receive size.
	*/
	iip->iic_mrblr = CPM_MAX_XFER;
}

static void
cpm_iic_print_params(struct i2c_algo_cpm_data *cpm)
{
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif
	volatile iic_t *iip = cpm->iip;

	printk("i2c regs @ 0x%08x:\n", (uint)i2c);
	printk("\tmod %02x add %02x brg %02x com %02x cer %02x cmr %02x\n",
		i2c->i2c_i2mod, i2c->i2c_i2add,
		i2c->i2c_i2brg, i2c->i2c_i2com,
		i2c->i2c_i2cer, i2c->i2c_i2cmr);

	printk("iic params @ 0x%08x:\n", (uint)iip);
	printk("\trbase %04x tbase %04x rfcr %02x tfcr %02x mrblr %04x\n",
		iip->iic_rbase, iip->iic_tbase,
		iip->iic_rfcr, iip->iic_tfcr,
		iip->iic_mrblr);
	printk("\trstate %08x rdp %08x rbptr %04x rbc %04x rxtmp %08x\n",
		iip->iic_rstate, iip->iic_rdp,
		iip->iic_rbptr, iip->iic_rbc,
		iip->iic_rxtmp);
	printk("\ttstate %08x tdp %08x tbptr %04x tbc %04x txtmp %08x\n",
		iip->iic_tstate, iip->iic_tdp,
		iip->iic_tbptr, iip->iic_tbc,
		iip->iic_txtmp);

#if defined(CONFIG_8xx)
	printk("\tres %08x rpbase %04x res2 %04x\n",
		iip->iic_res, iip->iic_rpbase, iip->iic_res2);
#elif defined(CONFIG_8260)
	printk("i2c base = %04x\n",
		*(ushort *)&immr->im_dprambase[PROFF_I2C_BASE]);
#endif

	if (using_reloc_patch(cpm))
		printk("Using relocation patch!\n");
}

static void
cpm_iic_init(struct i2c_algo_cpm_data *cpm)
{
#if defined(CONFIG_8xx)
	volatile i2c8xx_t	*i2c = cpm->i2c;
#elif defined(CONFIG_8260)
	volatile i2c8260_t	*i2c = cpm->i2c;
#endif

	if (cpm_iic_debug)
		printk("cpm_iic_init()\n");

	/* Disable interrupts.
	*/
	i2c->i2c_i2mod = I2C_DIS;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;

	/* Select an arbitrary address.  Just make sure it is unique.
	*/
	i2c->i2c_i2add = 0xfe;

	/* Make clock run maximum slow.
	*/
	i2c->i2c_i2brg = 7;

	/* Set I2C controller in master mode */
	i2c->i2c_i2com = 1;

	/* set up parameter ram */
	cpm_iic_reset_params(cpm);

	/* Initialize Tx/Rx parameters.
	*/
	if (!using_reloc_patch(cpm)) {
#if defined(CONFIG_8xx)
		volatile cpm8xx_t *cp = cpm->cp;
#elif defined(CONFIG_8260)
		volatile cpm8260_t *cp = cpm->cp;
#endif

		cp->cp_cpcr =
#if defined(CONFIG_8xx)
			mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_INIT_TRX)
#elif defined(CONFIG_8260)
			mk_cr_cmd(CPM_CR_I2C_PAGE, CPM_CR_I2C_SBLOCK,
				0, CPM_CR_INIT_TRX)
#endif
			| CPM_CR_FLG;

		while (cp->cp_cpcr & CPM_CR_FLG)
			;
	}

	/* Clear events and interrupts */
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;

	spin_lock_init(&data->buslock);

	init_waitqueue_head(&cpm->waitq);

	/* Install interrupt handler.
	*/
	if ((*cpm->setisr)(cpm, cpm_iic_interrupt))
		panic("i2c-algo-cpm.o: setisr failed");
}

static void
cpm_iic_force_close(struct i2c_algo_cpm_data *cpm)
{
	if (!using_reloc_patch(cpm)) {
#if defined(CONFIG_8xx)
		volatile cpm8xx_t *cp = cpm->cp;
#elif defined(CONFIG_8260)
		volatile cpm8260_t *cp = cpm->cp;
#endif

		if (cpm_iic_debug)
			printk("cpm_iic_force_close()\n");

		cp->cp_cpcr =
#if defined(CONFIG_8xx)
			mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_CLOSE_RXBD)
#elif defined(CONFIG_8260)
			mk_cr_cmd(CPM_CR_I2C_PAGE, CPM_CR_I2C_SBLOCK,
				0, CPM_CR_CLOSE_RXBD)
#endif
			| CPM_CR_FLG;

		while (cp->cp_cpcr & CPM_CR_FLG)
			;
	}
}

static int
cpm_iic_shutdown(struct i2c_algo_cpm_data *cpm)
{
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif

	/* Shut down IIC.
	*/
	i2c->i2c_i2mod = I2C_DIS;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;

	cpm_iic_force_close(cpm);

	(*cpm->setisr)(cpm, NULL);

	return(0);
}

#ifdef CONFIG_BROKEN_CPM_I2C
/*
 * got the following from Ian Abbott's driver - apparently
 * a NAK on a read could cause the CPM to lock up (I have
 * actually experienced this on an 8260 - Rev 01, Mask A.1
 * 1K22A-XC CPM@133MHz). Ian's driver does a loopback read
 * which is supposed to clear this condition.
 *
 * NOTE: must be called with the i2c bus locked
 */
static void
cpm_iic_loopback(struct i2c_algo_cpm_data *cpm)
{
	volatile iic_t *iip = cpm->iip;
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
	volatile cpm8xx_t *cp = cpm->cp;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif
	volatile cbd_t	*tbdf, *rbdf;
	u_long flags;
	long timeout;
	DECLARE_WAITQUEUE(wait, current);

	cpm->temp[0] = i2c->i2c_i2add;		/* write to our own address */
	cpm->temp[1] = 0;
	flush_dcache_range((u_long)cpm->temp, (u_long)(cpm->temp+2));

	/* check for and use a microcode relocation patch */
	if (using_reloc_patch(cpm))
		cpm_iic_reset_params(cpm);

#if defined(CONFIG_8xx)
	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];
#elif defined(CONFIG_8260)
	tbdf = (cbd_t *)&immr->im_dprambase[iip->iic_tbase];
	rbdf = (cbd_t *)&immr->im_dprambase[iip->iic_rbase];
#endif

	tbdf->cbd_bufaddr = __pa(cpm->temp);
	tbdf->cbd_datlen = 2;
	tbdf->cbd_sc = BD_SC_READY | BD_SC_LAST | BD_SC_WRAP | BD_IIC_START;

	rbdf->cbd_bufaddr = __pa(cpm->temp+2);
	rbdf->cbd_datlen = 0;
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT;

	/* Chip bug, set enable here */
	save_flags(flags); cli();
	i2c->i2c_i2cmr = 0x17;	/* Enable interrupts */
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2mod = I2C_EN;	/* Enable */
#if defined(CONFIG_8260)
	set_current_state(TASK_INTERRUPTIBLE);
	(void)schedule_timeout(1);	/* sleep for a jiffy */
#endif
	i2c->i2c_i2com = 0x81;	/* Start master */

	/* Wait for IIC transfer */
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&cpm->waitq, &wait);
	timeout = schedule_timeout(HZ/10);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&cpm->waitq, &wait);

	/* Chip errata, clear enable.
	*/
	i2c->i2c_i2mod = I2C_DIS;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;
	restore_flags(flags);

	if (signal_pending(current)) {
		if (cpm_iic_debug)
			printk("loopback sleep was interrupted!\n");
		return;
	}

	if (timeout == 0) {
		if (cpm_iic_debug)
			printk("loopback transfer timed-out!\n");
		return;
	}

	if (cpm_iic_debug)
		printk("tx sc %04x, rx sc %04x\n", tbdf->cbd_sc, rbdf->cbd_sc);

	if (tbdf->cbd_sc & BD_SC_XER) {
		if (cpm_iic_debug)
			printk("IIC loopback; xmit err (%x)\n",
				tbdf->cbd_sc & BD_SC_XER);
		return;
	}

	if (rbdf->cbd_sc & BD_SC_OV) {
		if (cpm_iic_debug)
			printk("IIC loopback; overrun\n");
		return;
	}

	if (rbdf->cbd_sc & BD_SC_EMPTY) {
		if (cpm_iic_debug)
			printk("IIC loopback; complete but rbuf empty\n");
		cpm_iic_force_close(cpm);
		if (cpm_iic_debug)
			printk("tx sc %04x, rx sc %04x\n",
			       tbdf->cbd_sc, rbdf->cbd_sc);
	}

	if (cpm_iic_debug)
		printk("loopback read %d bytes\n", rbdf->cbd_datlen);
}
#endif

/* Read from IIC...
 * abyte = address byte, with r/w flag already set
 */
static int
cpm_iic_read(struct i2c_algo_cpm_data *cpm, u_char abyte, char *buf, int count)
{
	volatile iic_t *iip = cpm->iip;
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
	volatile cpm8xx_t *cp = cpm->cp;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif
	volatile cbd_t	*tbdf, *rbdf;
	u_long flags;
	long timeout;
	DECLARE_WAITQUEUE(wait, current);

	if (count >= CPM_MAX_XFER)
		return -EINVAL;

	if (cpm_iic_debug)
		printk("cpm_iic_read(abyte=0x%x)\n", abyte);

	/* prevent simultaneous access to the i2c h/w */
	spin_lock(&cpm->buslock);

	/* check for and use a microcode relocation patch */
	if (using_reloc_patch(cpm))
		cpm_iic_reset_params(cpm);

#if defined(CONFIG_8xx)
	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];
#elif defined(CONFIG_8260)
	tbdf = (cbd_t *)&immr->im_dprambase[iip->iic_tbase];
	rbdf = (cbd_t *)&immr->im_dprambase[iip->iic_rbase];
#endif

	/* To read, we need an empty buffer of the proper length.
	 * All that is used is the first byte for address, the remainder
	 * is just used for timing (and doesn't really have to exist).
	 */
	cpm->temp[0] = abyte;		/* Device address byte w/rw flag */
	flush_dcache_range((u_long)cpm->temp, (u_long)(cpm->temp+1));

	tbdf->cbd_bufaddr = __pa(cpm->temp);
	tbdf->cbd_datlen = 1 + count;
	tbdf->cbd_sc = BD_SC_READY | BD_SC_LAST | BD_SC_WRAP | BD_IIC_START;

	rbdf->cbd_datlen = 0;
	rbdf->cbd_bufaddr = __pa(buf);
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT;

	/* Chip bug, set enable here */
	save_flags(flags); cli();
	i2c->i2c_i2cmr = 0x17;	/* Enable interrupts */
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2mod = I2C_EN;	/* Enable */
#if defined(CONFIG_8260)
	set_current_state(TASK_INTERRUPTIBLE);
	(void)schedule_timeout(1);	/* sleep for a jiffy */
#endif
	i2c->i2c_i2com = 0x81;	/* Start master */

	/* Wait for IIC transfer */
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&cpm->waitq, &wait);
	timeout = schedule_timeout(HZ/10);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&cpm->waitq, &wait);

	/* Chip errata, clear enable.
	*/
	i2c->i2c_i2mod = I2C_DIS;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;
	restore_flags(flags);

	if (signal_pending(current)) {
		if (cpm_iic_debug)
			printk("sleep was interrupted!\n");
		spin_unlock(&cpm->buslock);
		return -EINTR;
	}

	if (timeout == 0) {
		if (cpm_iic_debug)
			printk("transfer timed-out!\n");
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	if (cpm_iic_debug)
		printk("tx sc %04x, rx sc %04x\n", tbdf->cbd_sc, rbdf->cbd_sc);

	if (tbdf->cbd_sc & BD_SC_XER) {
		if (cpm_iic_debug)
			printk("IIC read; xmit err (%x)\n",
				tbdf->cbd_sc & BD_SC_XER);

#ifdef CONFIG_BROKEN_CPM_I2C
		/* try to stop hang after NAK'ed read */
		if (tbdf->cbd_sc & BD_SC_NAK)
			cpm_iic_loopback(cpm);
#endif

		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	if (rbdf->cbd_sc & BD_SC_OV) {
		if (cpm_iic_debug)
			printk("IIC read; overrun\n");
		spin_unlock(&cpm->buslock);
		return -EIO;
	}

	if (rbdf->cbd_sc & BD_SC_EMPTY) {
		if (cpm_iic_debug)
			printk("IIC read; complete but rbuf empty\n");
		cpm_iic_force_close(cpm);
		if (cpm_iic_debug)
			printk("tx sc %04x, rx sc %04x\n",
			       tbdf->cbd_sc, rbdf->cbd_sc);
	}

	if (cpm_iic_debug)
		printk("read %d bytes\n", rbdf->cbd_datlen);

	if (rbdf->cbd_datlen < count) {
		if (cpm_iic_debug)
			printk("IIC read; short, wanted %d got %d\n",
			       count, rbdf->cbd_datlen);
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	invalidate_dcache_range((u_long)buf, (u_long)(buf+count));

	spin_unlock(&cpm->buslock);
	return count;
}

/* Write to IIC...
 * addr = address byte, with r/w flag already set
 */
static int
cpm_iic_write(struct i2c_algo_cpm_data *cpm, u_char abyte, char *buf, int count)
{
	volatile iic_t *iip = cpm->iip;
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
	volatile cpm8xx_t *cp = cpm->cp;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif
	volatile cbd_t	*tbdf;
	u_long flags;
	long timeout;
	DECLARE_WAITQUEUE(wait, current);

	if (count >= CPM_MAX_XFER)
		return -EINVAL;

	if (cpm_iic_debug)
		printk("cpm_iic_write(abyte=0x%x)\n", abyte);

	/* prevent simultaneous access to the i2c h/w */
	spin_lock(&cpm->buslock);

	/* it looks to me like back-to-back i2c xfers simply don't work,
	 * at least on my MPC8260 cpu (Rev 01, Mask A.1 1K22A-XC CPM@133MHz).
	 * so can't use two transmit bd's, one for address and one for the
	 * data passed in (as in the original code). instead I copy it all
	 * into the temp buffer - i2c messages are usually small anyway, so
	 * the extra copy won't matter much, I reckon.
	 */
	cpm->temp[0] = abyte;		/* Device address byte w/rw flag */
	memcpy(&cpm->temp[1], buf, count);
	flush_dcache_range((u_long)cpm->temp, (u_long)(cpm->temp+count+1));

	/* check for and use a microcode relocation patch */
	if (using_reloc_patch(cpm))
		cpm_iic_reset_params(cpm);

	if (cpm_iic_debug)
		cpm_iic_print_params(cpm);

	/* set up descriptor */
#if defined(CONFIG_8xx)
	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
#elif defined(CONFIG_8260)
	tbdf = (cbd_t *)&immr->im_dprambase[iip->iic_tbase];
#endif
	tbdf->cbd_bufaddr = __pa(cpm->temp);
	tbdf->cbd_datlen = 1 + count;
	tbdf->cbd_sc = BD_SC_READY | BD_SC_INTRPT | BD_SC_START | BD_SC_LAST | BD_SC_WRAP;

	if (cpm_iic_debug)
		printk("tx sc %04x dl %04x ba %08x\n",
		       tbdf->cbd_sc, tbdf->cbd_datlen, tbdf->cbd_bufaddr);

	/* Chip bug, set enable here */
	save_flags(flags); cli();
	i2c->i2c_i2cmr = 0x17;	/* Enable interrupts */
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2mod = I2C_EN;	/* Enable */
#if defined(CONFIG_8260)
	set_current_state(TASK_INTERRUPTIBLE);
	(void)schedule_timeout(1);	/* sleep for a jiffy */
#endif
	i2c->i2c_i2com = 0x81;	/* Start master */

	/* Wait for IIC transfer */
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&cpm->waitq, &wait);
	timeout = schedule_timeout(HZ/10);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&cpm->waitq, &wait);

	/* Chip errata, clear enable.
	*/
	i2c->i2c_i2mod = I2C_DIS;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;
	restore_flags(flags);

	if (cpm_iic_debug) {
		cpm_iic_print_params(cpm);
		printk("tx sc %04x dl %04x ba %08x\n",
		       tbdf->cbd_sc, tbdf->cbd_datlen, tbdf->cbd_bufaddr);
	}

	if (signal_pending(current)) {
		if (cpm_iic_debug)
			printk("sleep was interrupted!\n");
		spin_unlock(&cpm->buslock);
		return -EINTR;
	}

	if (timeout == 0) {
		if (cpm_iic_debug)
			printk("transfer timed-out!\n");
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	if (tbdf->cbd_sc & BD_SC_XER) {
		if (cpm_iic_debug)
			printk("IIC write; xmit err %x\n",
				tbdf->cbd_sc & BD_SC_XER);
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	if (tbdf->cbd_sc & BD_SC_READY) {
		if (cpm_iic_debug)
			printk("IIC write; complete but tbd rdy\n");
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	spin_unlock(&cpm->buslock);
	return count;
}

static void printbd(volatile cbd_t *bd, u_char *va)
{
    int i;
    printk("BD @ %08lx: pa %08lx va %08lx len %04lx, flags %04lx\n",
	   (u_long)bd, (u_long)bd->cbd_bufaddr, (u_long)va, 
	   (u_long)bd->cbd_datlen, (u_long)bd->cbd_sc);
    printk("buffer: ");
    for (i=0; i<bd->cbd_datlen; i++)
      printk("%02x ", va[i]);
    printk("\n");
}

/* Write followed by a read, without STOP in between ...
 * addr = address byte, with r/w flag already set
 */
static int
cpm_iic_wrrd(struct i2c_algo_cpm_data *cpm,
	u_char tabyte, char *tbuf, int tcount,
	u_char rabyte, char *rbuf, int rcount)
{
	volatile iic_t *iip = cpm->iip;
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
	volatile cpm8xx_t *cp = cpm->cp;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif
	volatile cbd_t	*tbdf, *rbdf;
	u_long flags;
	long timeout;
	DECLARE_WAITQUEUE(wait, current);

	if (tcount >= CPM_MAX_XFER || rcount >= CPM_MAX_XFER)
		return -EINVAL;

	if (cpm_iic_debug)
		printk("cpm_iic_wrrd(tabyte=0x%x, rabyte=0x%x)\n",
			tabyte, rabyte);

	/* prevent simultaneous access to the i2c h/w */
	spin_lock(&cpm->buslock);

	cpm->temp[0] = tabyte;		/* Device address byte w/rw flag */
	memcpy(&cpm->temp[1], tbuf, tcount);
	cpm->temp[1+tcount] = rabyte;	/* Device address byte w/rw flag */
	flush_dcache_range((u_long)cpm->temp, (u_long)(cpm->temp+1+tcount+1));

	/* check for and use a microcode relocation patch */
	if (using_reloc_patch(cpm))
		cpm_iic_reset_params(cpm);

	if (cpm_iic_debug)
		cpm_iic_print_params(cpm);

	/* set up 3 descriptors */
#if defined(CONFIG_8xx)
	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];
#elif defined(CONFIG_8260)
	tbdf = (cbd_t *)&immr->im_dprambase[iip->iic_tbase];
	rbdf = (cbd_t *)&immr->im_dprambase[iip->iic_rbase];
#endif

	tbdf[0].cbd_bufaddr = __pa(cpm->temp);
	tbdf[0].cbd_datlen = 1 + tcount;
	tbdf[0].cbd_sc = BD_SC_READY | BD_IIC_START; /* no BD_SC_LAST no STOP */

	/* the rcount is only for timing - only one byte is transmitted
	 * because the device will recognise transmission of a read address
	 * and turn the line around - it will only do this if START flag is
	 * set though. Also because it begins reading, the LAST flag won't
	 * cause a STOP to be sent (I'm guessing about all this :-)
	 */
	tbdf[1].cbd_bufaddr = __pa(cpm->temp + 1 + tcount);
	tbdf[1].cbd_datlen = 1 + rcount;
	tbdf[1].cbd_sc = BD_SC_READY | BD_IIC_START | BD_SC_LAST | BD_SC_WRAP;

	rbdf->cbd_datlen = 0;
	rbdf->cbd_bufaddr = __pa(rbuf);
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT;

	if (cpm_iic_debug) {
		printbd(&tbdf[0], cpm->temp);
		printbd(&tbdf[1], cpm->temp + 1 + tcount);
		printbd(rbdf, rbuf);
	}

	/* Chip bug, set enable here */
	save_flags(flags); cli();
	i2c->i2c_i2cmr = 0x17;	/* Enable interrupts */
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2mod = I2C_EN;	/* Enable */
#if defined(CONFIG_8260)
	set_current_state(TASK_INTERRUPTIBLE);
	(void)schedule_timeout(1);	/* sleep for a jiffy */
#endif
	i2c->i2c_i2com = 0x81;	/* Start master */

	/* Wait for IIC transfer */
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&cpm->waitq, &wait);
	timeout = schedule_timeout(HZ/10);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&cpm->waitq, &wait);
   
	if (cpm_iic_debug) {
		printbd(&tbdf[0], cpm->temp);
		printbd(&tbdf[1], cpm->temp + 1 + tcount);
		invalidate_dcache_range((u_long)rbuf, (u_long)(rbuf+rcount));
		printbd(rbdf, rbuf);
	}

	/* Chip errata, clear enable.
	*/
	i2c->i2c_i2mod = I2C_DIS;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;
	restore_flags(flags);

	if (signal_pending(current)) {
		if (cpm_iic_debug)
			printk("sleep was interrupted!\n");
		spin_unlock(&cpm->buslock);
		return -EINTR;
	}

	if (timeout == 0) {
		if (cpm_iic_debug)
			printk("transfer timed-out!\n");
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	if (cpm_iic_debug) {
		cpm_iic_print_params(cpm);
		printk("tx0 sc %04x, tx1 sc %04x, rx sc %04x\n",
		       tbdf[0].cbd_sc, tbdf[1].cbd_sc, rbdf->cbd_sc);
	}

	if ((tbdf[0].cbd_sc & BD_SC_XER) || (tbdf[1].cbd_sc & BD_SC_XER)) {
		if (cpm_iic_debug)
			printk("IIC wrrd; xmit err 0=%x, 1=%x\n",
				tbdf[0].cbd_sc & BD_SC_XER,
				tbdf[1].cbd_sc & BD_SC_XER);

#ifdef CONFIG_BROKEN_CPM_I2C
		/* try to stop hang after NAK'ed read */
		if (tbdf[1].cbd_sc & BD_SC_NAK)
			cpm_iic_loopback(cpm);
#endif

		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	if ((tbdf[0].cbd_sc & BD_SC_READY) || (tbdf[1].cbd_sc & BD_SC_READY)) {
		if (cpm_iic_debug)
			printk("IIC wrrd; complete but tbuf rdy (%d)\n",
				(tbdf[0].cbd_sc & BD_SC_READY) ? 0 : 1);
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	if (rbdf->cbd_sc & BD_SC_OV) {
		if (cpm_iic_debug)
			printk("IIC wrrd; read overrun\n");
		spin_unlock(&cpm->buslock);
		return -EIO;
	}

	if (rbdf->cbd_sc & BD_SC_EMPTY) {
		if (cpm_iic_debug)
			printk("IIC read; complete but rbuf empty\n");
		cpm_iic_force_close(cpm);
		if (cpm_iic_debug)
			printk("tx0 sc %04x, tx1 sc %04x, rx sc %04x\n",
			       tbdf[0].cbd_sc, tbdf[1].cbd_sc, rbdf->cbd_sc);
	}

	if (cpm_iic_debug)
		printk("wrrd %d bytes\n", rbdf->cbd_datlen);

	if (rbdf->cbd_datlen < rcount) {
		if (cpm_iic_debug)
			printk("IIC wrrd; short read, wanted %d got %d\n",
			       rcount, rbdf->cbd_datlen);
		spin_unlock(&cpm->buslock);
		return -EREMOTEIO;
	}

	invalidate_dcache_range((u_long)rbuf, (u_long)(rbuf+rcount));

	spin_unlock(&cpm->buslock);
	return rcount;
}

/* See if an IIC address exists..
 * addr = 7 bit address, unshifted
 */
static int
cpm_iic_tryaddress(struct i2c_algo_cpm_data *cpm, int addr)
{
	volatile iic_t *iip = cpm->iip;
#if defined(CONFIG_8xx)
	volatile i2c8xx_t *i2c = cpm->i2c;
	volatile cpm8xx_t *cp = cpm->cp;
#elif defined(CONFIG_8260)
	volatile i2c8260_t *i2c = cpm->i2c;
#endif
	volatile cbd_t *tbdf, *rbdf;
	u_long flags;
	long timeout;
	DECLARE_WAITQUEUE(wait, current);

	if (cpm_iic_debug > 1)
		printk("cpm_iic_tryaddress(cpm=%p,addr=%d)\n", cpm, addr);

	if (cpm_iic_debug && addr == 0) {
		printk("iip %p, iic_rbase 0x%x, dp_rbase 0x%x\n",
			cpm->iip, iip->iic_rbase, cpm->dp_rbase);
		printk("iic_tbase %d, dp_tbase %d\n",
			iip->iic_tbase, cpm->dp_tbase);
	}

	/* prevent simultaneous access to the i2c h/w */
	spin_lock(&cpm->buslock);

	/* do a simple read */
	cpm->temp[0] = (addr << 1) | 1;	/* device address (+ read) */
	flush_dcache_range((u_long)cpm->temp, (u_long)(cpm->temp+1));

	/* check for and use a microcode relocation patch */
	if (using_reloc_patch(cpm))
		cpm_iic_reset_params(cpm);

#if defined(CONFIG_8xx)
	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];
#elif defined(CONFIG_8260)
	tbdf = (cbd_t *)&immr->im_dprambase[iip->iic_tbase];
	rbdf = (cbd_t *)&immr->im_dprambase[iip->iic_rbase];
#endif

	tbdf->cbd_bufaddr = __pa(cpm->temp);
	tbdf->cbd_datlen = 2;
	tbdf->cbd_sc = BD_SC_READY | BD_SC_LAST | BD_SC_WRAP | BD_IIC_START;

	rbdf->cbd_datlen = 0;
	rbdf->cbd_bufaddr = __pa(cpm->temp+2);
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP | BD_SC_INTRPT;

	/* Chip bug, set enable here */
	save_flags(flags); cli();
	i2c->i2c_i2cmr = 0x17;	/* Enable interrupts */
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2mod = I2C_EN;	/* Enable */
#if defined(CONFIG_8260)
	(void)schedule_timeout(1);	/* sleep for a jiffy */
#endif
	i2c->i2c_i2com = 0x81;	/* Start master */

	if (cpm_iic_debug > 1)
		printk("about to sleep\n");

	/* wait for IIC transfer */
	set_current_state(TASK_INTERRUPTIBLE);
	add_wait_queue(&cpm->waitq, &wait);
	timeout = schedule_timeout(HZ/10);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&cpm->waitq, &wait);

	/* Chip errata, clear enable.
	*/
	i2c->i2c_i2mod = I2C_DIS;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;
	restore_flags(flags);

	if (cpm_iic_debug > 1)
		printk("back from sleep\n");

	if (signal_pending(current)) {
		if (cpm_iic_debug)
			printk("sleep was interrupted!\n");
		spin_unlock(&cpm->buslock);
		return -1;
	}

	if (timeout == 0) {
		if (cpm_iic_debug)
			printk("transfer timed-out!\n");
		spin_unlock(&cpm->buslock);
		return 0;
	}

	if (tbdf->cbd_sc & BD_SC_XER) {
		if (cpm_iic_debug > 1)
			printk("IIC try; xmit err (%x)\n",
				tbdf->cbd_sc & BD_SC_XER);

#ifdef CONFIG_BROKEN_CPM_I2C
		/* try to stop hang after NAK'ed read */
		if (tbdf->cbd_sc & BD_SC_NAK)
			cpm_iic_loopback(cpm);
#endif

		spin_unlock(&cpm->buslock);
		return 0;
	}

	if (tbdf->cbd_sc & BD_SC_READY) {
		printk("IIC try; complete but tbuf ready\n");
	}

	spin_unlock(&cpm->buslock);
	return 1;
}

static int
cpm_iic_xfer(struct i2c_adapter *i2c_adap, struct i2c_msg msgs[], int num)
{
	struct i2c_algo_cpm_data *adap = i2c_adap->algo_data;
	struct i2c_msg *pmsg, *pmsgn;
	int i, ret;
	u_char addr;

	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];
		pmsgn = (i < num - 1) ? &msgs[i + 1] : NULL;

		if (cpm_iic_debug)
			printk("i2c-algo-cpm.o: "
			       "#%d addr=0x%x flags=0x%x len=%d\n",
			       i, pmsg->addr, pmsg->flags, pmsg->len);
    
		if (pmsg->flags & I2C_M_NOSTART) {
			if (cpm_iic_debug)
				printk("i2c-algo-cpm.o: NOSTART not impl!\n");
			return -EINVAL;
		}

		addr = pmsg->addr << 1;
		if (pmsg->flags & I2C_M_RD )
			addr |= 1;
		if (pmsg->flags & I2C_M_REV_DIR_ADDR )
			addr ^= 1;

		if (pmsg->flags & I2C_M_RD ) {
			/* read bytes into buffer*/
			ret = cpm_iic_read(adap, addr, pmsg->buf, pmsg->len);
			if (cpm_iic_debug)
				printk("i2c-algo-cpm.o: read %d\n", ret);
			if (ret < pmsg->len ) {
				return (ret<0)? ret : -EREMOTEIO;
			}
		} else {
			if (pmsgn != NULL &&
			    (pmsgn->flags & I2C_M_RD) &&
			    !(pmsgn->flags & I2C_M_NOSTART) &&
			    pmsgn->addr == pmsg->addr) {
				u_char rdaddr;

				rdaddr = (pmsgn->addr << 1) | 1;
				if (pmsgn->flags & I2C_M_REV_DIR_ADDR)
					rdaddr ^= 1;

				/* write, then read, without STOP in between */
				ret = cpm_iic_wrrd(adap,
					addr, pmsg->buf, pmsg->len,
					rdaddr, pmsgn->buf, pmsgn->len);
				if (cpm_iic_debug)
					printk("i2c-algo-cpm.o: wrrd %d\n",
						ret);
				if (ret <= 0) {
					return (ret<0) ? ret : -EREMOTEIO;
				}

				i++;	/* this transaction did two msgs */
			} else {
				/* write bytes from buffer */
				ret = cpm_iic_write(adap,
					addr, pmsg->buf, pmsg->len);
				if (cpm_iic_debug)
					printk("i2c-algo-cpm.o: wrote %d\n",
						ret);
				if (ret < pmsg->len ) {
					return (ret<0) ? ret : -EREMOTEIO;
				}
			}
		}
	}
	return (num);
}

static int
cpm_iic_ctrl(struct i2c_adapter *adapter, unsigned int cmd, unsigned long arg)
{
	return 0;
}

static u32
cpm_iic_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_SMBUS_EMUL | I2C_FUNC_PROTOCOL_MANGLING; 
}

/* -----exported algorithm data: -------------------------------------	*/

static struct i2c_algorithm cpm_algo = {
	name:		"MPC8XX/8260 CPM algorithm",
	id:		I2C_ALGO_CPM,
	master_xfer:	cpm_iic_xfer,
	algo_control:	cpm_iic_ctrl,
	functionality:	cpm_iic_func,
};

/* 
 * registering functions to load algorithms at runtime 
 */
int
i2c_cpm_add_bus(struct i2c_adapter *adap)
{
	int i;
	struct i2c_algo_cpm_data *cpm = adap->algo_data;

	if (cpm_iic_debug)
		printk("i2c-algo-cpm.o: hw routines for %s registered.\n",
		       adap->name);

	/* register new adapter to i2c module... */

	adap->id |= cpm_algo.id;
	adap->algo = &cpm_algo;

	MOD_INC_USE_COUNT;

	i2c_add_adapter(adap);
	cpm_iic_init(cpm);

	/* scan bus */
	if (cpm_iic_scan) {
		printk(KERN_INFO " i2c-algo-cpm.o: scanning bus %s ...\n",
		       adap->name);
		for (i = 0; i < 128; i++) {
			int n = cpm_iic_tryaddress(cpm, i);

			if (n < 0) {
				printk("interrupted!");
				break;
			}

			if (n > 0) {
				printk("(%02x)",i<<1); 
			}
		}
		printk("\n");
	}
	return 0;
}

int
i2c_cpm_del_bus(struct i2c_adapter *adap)
{
	int res;
	struct i2c_algo_cpm_data *cpm = adap->algo_data;

	cpm_iic_shutdown(cpm);

	if ((res = i2c_del_adapter(adap)) < 0)
		return res;

	printk("i2c-algo-cpm.o: adapter unregistered: %s\n",adap->name);
	MOD_DEC_USE_COUNT;

	return 0;
}

EXPORT_SYMBOL(i2c_cpm_add_bus);
EXPORT_SYMBOL(i2c_cpm_del_bus);

int __init
i2c_algo_cpm_init(void)
{
	printk("i2c-algo-cpm.o: i2c mpc8xx/8260 CPM algorithm module\n");
	return 0;
}

#ifdef MODULE
MODULE_AUTHOR("Brad Parker <brad@heeltoe.com>");
MODULE_DESCRIPTION("I2C-Bus MPC8XX/8260 CPM algorithm");

MODULE_PARM(cpm_iic_scan, "i");
MODULE_PARM_DESC(cpm_iic_scan, "Scan i2c bus on init");

MODULE_PARM(cpm_iic_debug, "i");
MODULE_PARM_DESC(cpm_iic_debug, "Debug level");

int
init_module(void) 
{
	return i2c_algo_cpm_init();
}

void
cleanup_module(void) 
{
	return;
}
#endif
