/*
 * Embedded Planet RPX Lite MPC8xx CPM I2C interface.
 * Copyright (c) 1999 Dan Malek (dmalek@jlc.net).
 *
 * moved into proper i2c interface;
 * Brad Parker (brad@heeltoe.com)
 *
 * RPX lite specific parts of the i2c interface
 *
 * Hacked to provide a generic CPM i2c driver for
 * either 8xx or 8260; renamed i2c-cpm.c
 * by Murray.Jensen@cmst.csiro.au, 17-May-01
 */

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/stddef.h>
#include <linux/parport.h>

#if defined(CONFIG_8xx)
#include <asm/mpc8xx.h>
#include "../../arch/ppc/8xx_io/commproc.h"
#elif defined(CONFIG_8260)
#include <asm/mpc8260.h>
#include <asm/cpm_8260.h>
#include <asm/iopin_8260.h>
#include <asm/irq.h>
#endif

#include <linux/i2c.h>
#include <linux/i2c-algo-cpm.h>

#if defined(CONFIG_8260)
static iopin_t scl_pin = { IOPIN_PORTD, 14 };	/* SCL pin is PD14 */
static iopin_t sda_pin = { IOPIN_PORTD, 15 };	/* SDA pin is PD15 */
#endif

static void
cpm_iic_init(struct i2c_algo_cpm_data *data)
{
	uint dpaddr;
	int count;
#if defined(CONFIG_8xx)
	volatile cpm8xx_t *cp;
	volatile immap_t *immap;

	cp = cpmp;	/* Get pointer to Communication Processor */
	immap = (immap_t *)IMAP_ADDR;	/* and to internal registers */

	data->iip = (iic_t *)&cp->cp_dparam[PROFF_IIC];

	/* Check for and use a microcode relocation patch.
	*/
	if ((data->reloc = data->iip->iic_rpbase))
		data->iip = (iic_t *)&cp->cp_dpmem[data->iip->iic_rpbase];
		
	data->i2c = (i2c8xx_t *)&(immap->im_i2c);
	data->cp = cp;

	/* Initialize Port B IIC pins.
	*/
	cp->cp_pbpar |= 0x00000030;
	cp->cp_pbdir |= 0x00000030;
	cp->cp_pbodr |= 0x00000030;

	/* Allocate space for two transmit and two receive buffer
	 * descriptors in the DP ram (if required).
	 *
	 * NOTE: these tests require initialisation to 0 in m8xx_cpm_reset()
	 */
	if ((dpaddr = data->iip->iic_tbase) == 0) {
		dpaddr = m8xx_cpm_dpalloc(sizeof (cbd_t) * 2);
		if (dpaddr == CPM_DP_NOSPACE)
			panic("no dpram available for i2c xmit bds");
		memset((void *)&cp->cp_dpmem[dpaddr], 0, sizeof (cbd_t) * 2);
		data->iip->iic_tbase = dpaddr;
	}
	data->dp_tbase = dpaddr;

	if ((dpaddr = data->iip->iic_rbase) == 0) {
		dpaddr = m8xx_cpm_dpalloc(sizeof (cbd_t) * 2);
		if (dpaddr == CPM_DP_NOSPACE)
			panic("no dpram available for i2c recv bds");
		memset((void *)&cp->cp_dpmem[dpaddr], 0, sizeof (cbd_t) * 2);
		data->iip->iic_rbase = dpaddr;
	}
	data->dp_rbase = dpaddr;
#elif defined(CONFIG_8260)
	volatile cpm8260_t *cp;
	volatile immap_t *immap;

	cp = cpmp;	/* Get pointer to Communication Processor */
	immap = immr;	/* and to internal registers */

	/* Check i2c parameter ram offset base in dual port ram. Allocate
	 * it if it hasn't been done before.
	 *
	 * NOTE: this test requires initialisation to 0 in m8260_cpm_reset()
	 */
	dpaddr = *(ushort *)&immr->im_dprambase[PROFF_I2C_BASE];
	if (dpaddr == 0) {
		/* need to allocate dpram for the i2c parameter block */
		dpaddr = m8260_cpm_dpalloc(sizeof (iic_t), 64);
		if (dpaddr == CPM_DP_NOSPACE)
			panic("no dpram available for i2c param block");
		memset((void *)&immr->im_dprambase[dpaddr], 0, sizeof (iic_t));
		*(ushort *)&immr->im_dprambase[PROFF_I2C_BASE] = dpaddr;
	}

	data->iip = (iic_t *)&immr->im_dprambase[dpaddr];

	data->i2c = (i2c8260_t *)&(immap->im_i2c);
	data->cp = cpmp;	/* Get pointer to Communication Processor */

	/* Initialize IIC pins.
	*/
	iopin_set_ded(&scl_pin);	/* set SCL pin as Dedicated ... */
	iopin_set_opt2(&scl_pin);	/* ... Option 2 ... */
	iopin_set_odr(&scl_pin);	/* ... Open Drain ... */
	iopin_set_in(&scl_pin);		/* ... Input (actually In/Out) */
	iopin_set_ded(&sda_pin);	/* set SDA pin as Dedicated ... */
	iopin_set_opt2(&sda_pin);	/* ... Option 2 ... */
	iopin_set_odr(&sda_pin);	/* ... Open Drain ... */
	iopin_set_in(&sda_pin);		/* ... Input (actually In/Out) */

	/* Allocate space for two transmit and two receive buffer
	 * descriptors in the DP ram (if required).
	 */
	if ((dpaddr = data->iip->iic_tbase) == 0) {
		dpaddr = m8260_cpm_dpalloc(sizeof (cbd_t) * 2, 8);
		if (dpaddr == CPM_DP_NOSPACE)
			panic("no dpram available for i2c xmit bds");
		memset((void *)&immr->im_dprambase[dpaddr], 0, sizeof(cbd_t)*2);
		data->iip->iic_tbase = dpaddr;
	}
	data->dp_tbase = dpaddr;

	if ((dpaddr = data->iip->iic_rbase) == 0) {
		dpaddr = m8260_cpm_dpalloc(sizeof (cbd_t) * 2, 8);
		if (dpaddr == CPM_DP_NOSPACE)
			panic("no dpram available for i2c recv bds");
		memset((void *)&immr->im_dprambase[dpaddr], 0, sizeof(cbd_t)*2);
		data->iip->iic_rbase = dpaddr;
	}
	data->dp_rbase = dpaddr;
#endif

	count = (CPM_MAX_XFER * 2) + 4;
	if ((data->temp = kmalloc(count, GFP_KERNEL)) == NULL)
		panic("i2c-cpm.o: kmalloc of %d bytes failed!\n", count);
}

static void
cpm_iic_shutdown(struct i2c_algo_cpm_data *data)
{
	if (data->temp != NULL) {
		kfree(data->temp);
		data->temp = NULL;
	}
}

#if defined(CONFIG_8260)
static void
cpm_call_isrfunc(int irq, void *data, struct pt_regs *regs)
{
	struct i2c_algo_cpm_data *cpm = (struct i2c_algo_cpm_data *)data;

	if (cpm->isrfunc != NULL)
		(*cpm->isrfunc)(data);
}
#endif

static int
cpm_install_isr(struct i2c_algo_cpm_data *cpm, void (*func)(void *))
{
	/* install interrupt handler */
#if defined(CONFIG_8xx)
	if (func == NULL)
		cpm_free_handler(CPMVEC_I2C);
	else
		cpm_install_handler(CPMVEC_I2C, func, (void *)cpm);
	return 0;
#elif defined(CONFIG_8260)
	if ((cpm->isrfunc = func) == NULL)
		return request_8xxirq(SIU_INT_I2C, NULL, 0, "i2c", (void *)cpm);
	else
		return request_8xxirq(SIU_INT_I2C, cpm_call_isrfunc,
			0, "i2c", (void *)cpm);
#endif
}

static int
cpm_iic_reg(struct i2c_client *client)
{
	return 0;
}

static int
cpm_iic_unreg(struct i2c_client *client)
{
	return 0;
}

static void
cpm_iic_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

static void
cpm_iic_dec_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

static struct i2c_algo_cpm_data cpm_iic_data = {
	setisr: cpm_install_isr
};

static struct i2c_adapter cpm_iic_ops = {
	"cpm",
	I2C_HW_MPC8XX_EPON,
	NULL,
	&cpm_iic_data,
	cpm_iic_inc_use,
	cpm_iic_dec_use,
	cpm_iic_reg,
	cpm_iic_unreg,
};

int __init
i2c_cpm_init(void)
{
	printk("i2c-cpm.o: i2c Generic CPM module\n");

	/* reset hardware to sane state */
	cpm_iic_init(&cpm_iic_data);

	if (i2c_cpm_add_bus(&cpm_iic_ops) < 0) {
		printk("i2c-cpm: Unable to register with I2C\n");
		return -ENODEV;
	}

	return 0;
}

void __exit i2c_cpm_exit(void)
{
	i2c_cpm_del_bus(&cpm_iic_ops);

	cpm_iic_shutdown(&cpm_iic_data);

	printk("i2c-cpm.o: i2c Generic CPM module removed\n");
}

#ifdef MODULE
MODULE_AUTHOR("Dan Malek <dmalek@jlc.net>");
MODULE_DESCRIPTION("I2C-Bus adapter routines for EP Generic CPM");

module_init(i2c_cpm_init);
module_exit(i2c_cpm_exit);
#endif
