/* ------------------------------------------------------------------------- */
/* i2c-men824xali.c i2c-hw access for MEN 824X ALI boards 					 */
/* used to access SMS24 on A12/D3/SC13/B11 boards							 */
/* ------------------------------------------------------------------------- */
/*   Copyright (C) 2002 Klaus Popp MEN Mikro Elektronik GmbH

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */

/* $Id: i2c-elv.c,v 1.17 2001/07/29 02:44:25 mds Exp $ */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>

#include <linux/errno.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <asm/pci-bridge.h>

#if !defined(CONFIG_MENB11) && !defined(CONFIG_MENA12)
# error "i2c-men824xali: Must define either CONFIG_MENB11 or CONFIG_MENA12"
#endif

#define ALI1543_PMU_DEV_NO      0x1C
#define ALI_DEVFN PCI_DEVFN(ALI1543_PMU_DEV_NO,0)

/* ----- global defines -----------------------------------------------	*/
#define DEB(x)		/* should be reasonable open, close &c. 	*/
#define DEB2(x) 	/* low level debugging - very slow 		*/
#define DEBE(x)	x	/* error messages 				*/
#define DEBINIT(x) x	/* detection status messages			*/

/* ALI PMU register offsets */
#define ALI1543_PMU_DOGPO1      0xc0   
#define ALI1543_PMU_DOGPO1_HMB  0xc2   
#define ALI1543_PMU_GPIO_DIR 	0x7d
#define ALI1543_PMU_GPIO_DOUT 	0x7e
#define ALI1543_PMU_GPIO_DIN 	0x7f

#ifdef CONFIG_MENB11
#  define SMB2_SCL			0x08	/* GPIO 3 */
#  define SMB2_SDA			0x80	/* GPIO 7 */
#endif

#ifdef CONFIG_MENA12
#  define SMB2_SCL			0x10	/* GPO 20 */
#  define SMB2_SDA			0x80	/* GPIO 7 */
#endif 

/* defined in arch/ppc/platforms/men824xali_pci.c */
extern struct pci_controller* men824xali_hose;

/* ----- local functions ----------------------------------------------	*/
static void set_gpio_low( u8 mask )
{
	u8 reg;

	/* program to low and output */
	early_read_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							ALI1543_PMU_GPIO_DOUT, &reg );
	early_write_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							 ALI1543_PMU_GPIO_DOUT, reg & ~mask );

	early_read_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							ALI1543_PMU_GPIO_DIR, &reg );
	early_write_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							 ALI1543_PMU_GPIO_DIR, reg | mask );	
}

static void set_gpio_high( u8 mask )
{
	u8 reg;

	/* program to input; pullups will raise it */
	early_read_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							ALI1543_PMU_GPIO_DIR, &reg );
	early_write_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							 ALI1543_PMU_GPIO_DIR, reg & ~mask );	
}

static void set_gpo_high( u8 mask )
{
	u8 reg;

	early_read_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							ALI1543_PMU_DOGPO1_HMB, &reg );
	early_write_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							 ALI1543_PMU_DOGPO1_HMB, reg | mask );	
}

static void set_gpo_low( u8 mask )
{
	u8 reg;

	early_read_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							ALI1543_PMU_DOGPO1_HMB, &reg );
	early_write_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							 ALI1543_PMU_DOGPO1_HMB, reg & ~mask );	
}


static int get_gpio( u8 mask )
{
	u8 reg;

	early_read_config_byte( men824xali_hose, 0, ALI_DEVFN, 
							ALI1543_PMU_GPIO_DIN, &reg );
	return (reg & mask) != 0;
}


static void bit_men824xali_setscl(void *data, int state)
{
#ifdef CONFIG_MENB11	
	if( state )
		set_gpio_high( SMB2_SCL );
	else
		set_gpio_low( SMB2_SCL );
#endif

#ifdef CONFIG_MENA12
	if( state )
		set_gpo_high( SMB2_SCL );
	else
		set_gpo_low( SMB2_SCL );
#endif
}

static void bit_men824xali_setsda(void *data, int state)
{
	if( state )
		set_gpio_high( SMB2_SDA );
	else
		set_gpio_low( SMB2_SDA );
} 

static int bit_men824xali_getscl(void *data)
{
	/* need to program to input before ??? */
#ifdef CONFIG_MENB11	
	return get_gpio( SMB2_SCL );
#endif
#ifdef CONFIG_MENA12
	return 1;					/* cannot read SCL on A12 */
#endif
}

static int bit_men824xali_getsda(void *data)
{
	return get_gpio( SMB2_SDA );	
}

static int bit_men824xali_init(void)
{
	bit_men824xali_setscl( NULL, 1);
	bit_men824xali_setsda( NULL, 1);
	return 0;
}

static void __exit bit_men824xali_exit(void)
{
}

static int bit_men824xali_reg(struct i2c_client *client)
{
	return 0;
}

static int bit_men824xali_unreg(struct i2c_client *client)
{
	return 0;
}

static void bit_men824xali_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

static void bit_men824xali_dec_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}

/* ------------------------------------------------------------------------
 * Encapsulate the above functions in the correct operations structure.
 * This is only done when more than one hardware adapter is supported.
 */
static struct i2c_algo_bit_data bit_men824xali_data = {
	NULL,
	bit_men824xali_setsda,
	bit_men824xali_setscl,
	bit_men824xali_getsda,
	bit_men824xali_getscl,
	80, 80, 100,		/*	waits, timeout */
};

static struct i2c_adapter bit_men824xali_ops = {
	"MEN824XALI 2nd SMB",
	I2C_HW_B_MEN824XALI,
	NULL,
	&bit_men824xali_data,
	bit_men824xali_inc_use,
	bit_men824xali_dec_use,
	bit_men824xali_reg,
	bit_men824xali_unreg,	
};

int __init i2c_bitmen824xali_init(void)
{
	printk("i2c-men824xali.o: i2c MEN824XALI 2nd SMB\n");

	if (bit_men824xali_init()==0) {
		if(i2c_bit_add_bus(&bit_men824xali_ops) < 0)
			return -ENODEV;
	} else {
		return -ENODEV;
	}
	printk("i2c-men824xali.o: init done\n");
	return 0;
}


EXPORT_NO_SYMBOLS;

#ifdef MODULE
MODULE_AUTHOR("Klaus Popp <klaus.popp@men.de>");
MODULE_DESCRIPTION("I2C-Bus adapter routines for MEN824XALI 2nd SMB");
MODULE_LICENSE("GPL");

int init_module(void)
{
	return i2c_bitmen824xali_init();
}

void cleanup_module(void)
{
	i2c_bit_del_bus(&bit_men824xali_ops);
	bit_men824xali_exit();
}

#endif

