/*
 *  linux/drivers/i2c/m41t56.c
 *
 *  Copyright (C) 2000 MontaVista Software, Inc.
 *
 *
 *  Copyright (C) 2000 Axlog Ingenierie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Driver for M41T56 TIMEKEEPER RTC & RAM chip
 */
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/mc146818rtc.h>
#include <linux/m41t56rtc.h>
#include <linux/init.h>

static struct i2c_driver m41t56_driver;

static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] = { (0xD0>>1), I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
	normal_i2c:		normal_addr,
	normal_i2c_range:	ignore,
	probe:			ignore,
	probe_range:		ignore,
	ignore:			ignore,
	ignore_range:		ignore,
	force:			ignore,
};


/* kmalloc not possible in early setup */
static struct i2c_client static_client = {
	name :	"00000000000000000000000000000000",
	id : 0,
	flags : 0,		
	addr : 0,						
	adapter : NULL,	
	driver : NULL,	
	data	: NULL,	
	usage_count : NULL,
};

#define DAT(x) ((unsigned int)(x->data))


static int
m41t56_attach(struct i2c_adapter *adap, int addr, unsigned short flags, int kind)
{
	struct i2c_client *c= &static_client;
	unsigned char buf[1];
	
	/*m41t56_init();*/

	/*c = kmalloc(sizeof(*c), GFP_KERNEL);*/
	

	if (!c)
		return -ENOMEM;

	strcpy(c->name, "M41T56");
	c->id		= m41t56_driver.id;
	c->flags	= flags|I2C_CLIENT_ALLOW_USE;
	c->addr		= addr;
	c->adapter	= adap;
	c->driver	= &m41t56_driver;
	c->data		= NULL;

	/* read control at the right device word address 7 */
	if ( (buf[0] = i2c_smbus_read_byte_data( c, 0  )) > 0)
	  DAT(c) = buf[0]; /* not NULL */

	return i2c_attach_client(c);
}

static int
m41t56_probe(struct i2c_adapter *adap)
{
#if 1
	return i2c_probe(adap, &addr_data, m41t56_attach);
#else
	return m41t56_attach(adap, 0xD0>>1, 0, 0);
#endif
}

static int
m41t56_detach(struct i2c_client *client)
{
	i2c_detach_client(client);

	return 0;
}

static int
m41t56_get_datetime(struct i2c_client *client, struct rtc_tm *dt)
{
	unsigned char buf[8];
	int ret = 0;
	int i;

	memset(buf, 0, sizeof(buf));

	/* read buf[i] at the right device word address i */
	for( i=0; i<8; i++)
	  {
	    ret = i2c_smbus_read_byte_data(client, i );
	    buf[i] = ret;

	    if (ret < 0)
	      break;
	    else
	      ret = 0;
	  }

	if (ret == 0) {
		buf[0] &= 0x7f; /* seconds */
		buf[1] &= 0x7f; /* minutes */
		buf[2] &= 0x3f; /* hours */
		buf[3] &= 0x07; /* day-of-week */
		buf[4] &= 0x3f; /* day-of-month */
		buf[5] &= 0x1f; /* month */
		buf[6] &= 0xff; /* year */

		dt->secs     = BCD_TO_BIN(buf[0]);
		dt->mins     = BCD_TO_BIN(buf[1]);
		dt->hours    = BCD_TO_BIN(buf[2]);
		dt->wday     = BCD_TO_BIN(buf[3]);
		dt->mday     = BCD_TO_BIN(buf[4]);
		dt->mon      = BCD_TO_BIN(buf[5]);
		dt->year     = BCD_TO_BIN(buf[6]);
		dt->cs       = BCD_TO_BIN(buf[7]);
	}

	return ret;
}

static int
m41t56_set_datetime(struct i2c_client *client, struct rtc_tm *dt, int datetoo)
{
	unsigned char buf[8];
	int ret, len = 3;
	int i;


	buf[0] = BIN_TO_BCD(dt->secs);
	buf[1] = BIN_TO_BCD(dt->mins);
	buf[2] = BIN_TO_BCD(dt->hours);

	if (datetoo) {
		len += 4;
		buf[3] = BIN_TO_BCD(dt->wday);
		buf[4] = BIN_TO_BCD(dt->mday);
		buf[5] = BIN_TO_BCD(dt->mon);
		buf[6] = BIN_TO_BCD(dt->year);
	}

	/* write buf[i] at the right device word address i */
	for( i=0; i<len; i++)
	  {
	    ret = i2c_smbus_write_byte_data(client, i, buf[i] );
	    
	    if (ret != 0)
	      break;
	  }

	return ret;
}

static int
m41t56_get_ctrl(struct i2c_client *client, unsigned char *ctrl)
{
	*ctrl = DAT(client);
	return 0;
}

static int
m41t56_set_ctrl(struct i2c_client *client, unsigned char *ctrl)
{
	DAT(client) = *ctrl;

	/* write control at word address 7 */
	return i2c_smbus_write_byte_data(client, 7, *ctrl );
}

static int
m41t56_read_mem(struct i2c_client *client, struct mem *mem)
{
	int i, j;
	int ret = 0;

	/* memory starts at address 8 */
	if (mem->loc < 8)
		return -EINVAL;

	/* read data[i] at the right device word address j */
	for( i=0, j=mem->loc; i<mem->nr; i++, j++)
	  {
	    /* word address < 64 */
	    if( j > 63)
	      {
		ret = -1;
		break;
	      }

	    ret = i2c_smbus_read_byte_data(client, j );
	    mem->data[i] = ret;
	    
	    if (ret < 0)
	      break;
	    else
	      ret = 0;
	  }
	return ret;
}

static int
m41t56_write_mem(struct i2c_client *client, struct mem *mem)
{
	int i, j, ret;

	/* memory starts at address 8 */
	if (mem->loc < 8)
		return -EINVAL;

	/* write data[i] at the right device word address j */
	for( i=0, j=mem->loc; i<mem->nr; i++, j++)
	  {
	    /* word address < 64 */
	    if( j > 63)
	      {
		ret = -1;
		break;
	      }

	    ret = i2c_smbus_write_byte_data(client, j, mem->data[i] );
	    
	    if (ret != 0)
	      break;
	  }
	return ret;
}

static int
m41t56_command(struct i2c_client *client, unsigned int cmd, void *arg)
{
	switch (cmd) {
	case RTC_GETDATETIME:
		return m41t56_get_datetime(client, arg);
		
	case RTC_SETTIME:
		return m41t56_set_datetime(client, arg, 0);

	case RTC_SETDATETIME:
		return m41t56_set_datetime(client, arg, 1);

	case RTC_GETCTRL:
		return m41t56_get_ctrl(client, arg);

	case RTC_SETCTRL:
		return m41t56_set_ctrl(client, arg);

	case MEM_READ:
		return m41t56_read_mem(client, arg);

	case MEM_WRITE:
		return m41t56_write_mem(client, arg);

	default:
		return -EINVAL;
	}
}

static struct i2c_driver m41t56_driver = {
	name:		"M41T56",
	id:		I2C_DRIVERID_M41T56,
	flags:		I2C_DF_NOTIFY,
	attach_adapter:	m41t56_probe,
	detach_client:	m41t56_detach,
	command:	m41t56_command
};

/*static*/ __init int m41t56_init(void)
{
	return i2c_add_driver(&m41t56_driver);
}

#ifdef MODULE
__initcall(m41t56_init);
#endif
